from __future__ import print_function
"""
-------------------------------
(c) Universite catholique de Louvain, 2019

Creation : 2019 by O. Lantsoght
Last update : 2019

version MBsysC v1.11.2

-------------------------------
Portable Python interface to MBsysC using Ctypes.


"""

import numpy as np
# importing libraries
from ._mbsysc_loader.loadlibs import libutilities

# importing wrapping function
from ._mbsysc_loader.callback import mbs_messages_wrap

# Function definition

def bytes_to_str(message):
    """
    Function converting bytes to str using 'utf-8' encoding.
    
    Strings retrived from C-libraries are array of bytes. In python3 (2) strings 
    type is 'str' ('unicode') with the 'utf-8' encoding. The conversion from
    bytes to str is required to get nice printing message or to compare string
    from C-libraries to string provided by the user from Python.
    
    Parameters
    ----------
    message: str or bytes
        Message to be converted, if needed, in 'str' type.
    
    Returns
    -------
        The same message provided byt as a 'str' type with 'utf-8' encoding
    """
    if type(message) is bytes:
        return message.decode("utf-8")
    
    return message

def str_to_bytes(message):
    """
    Function converting str to bytes using 'utf-8' encoding.
    
    In python3 strings type is 'str' while the memory in the C-libraries is 
    an array of bytes. When setting a string in the C-memory from a string defined
    in Python we have to ensure to provide a string that have the type 'bytes'.
    
    In python2 the type of strings may be 'str' or 'unicode'. A similar conversion
    is required.
    
    Parameters
    ----------
    message: str or unicode or bytes
        Message to be converted, if needed, in 'bytes' type (or 'str' compatible 
        with bytes in Python2).
    
    Returns
    -------
        The same message provided byt as a 'bytes' type with 'utf-8' encoding (or
        'str' compatible with bytes in Python2)
    """
    if type(message) is not bytes:
        return message.encode("utf-8")
    
    return message

def set_output(value, label):
    """
    Function to save a value during time integration.  
    
    Parameters
    ----------
    value : float
        Value to be saved.
    label : str
        The name of the output, also used as part as the output filename.
    """
    label_c = str_to_bytes(label)
    libutilities.set_output_ext(value, label_c)
    
def callback_undefined(function_name=None):
    """ 
    Internal function called when a unloaded function is called.
    
    This is required as calling a function who pointer is 'None' (or 'NULL' in 
    C) leads to an undefined behavior.    
    
    Parameters
    ----------
    function_name : str
        The name of the function that was called.
    
    """
    print("undefined function: "+function_name)
    return

def define_output_vector(label, size):
    """
    Function initializing the memory for a vector as output.
    
    This function must be called in the user init function dedicated to an
    analysis module.
    The value are defined during the computation thank to the function 
    'set_output_vector()' or 'set_output_value()'
    
    Parameters
    ----------
    label : str
        The name of the output. It is used as part of the filename and as 
        identifier when filling the values in the output.
    size : int
        The size of the output vector, which correspond to the maximum value
        to be saved at each time step.
    """
    label_c = str_to_bytes(label)
    libutilities.define_output_vector_ext(label_c, size)

def set_output_vector(vec, label):
    """
    Function to send an entire vector as output.
    
    The size of 'vec' must be equal to the size defined in the function 
    'define_output_vector()'.    
    
    Parameters
    ----------
    vec : list or numpy.ndarray
        Value to be saved as a list or a numpy array.
    label : str
        The name of the output. It must have been defined with the function 
    'define_output_vector()'.
    """
    label_c = str_to_bytes(label)
    for i in range(np.size(vec)):
        libutilities.set_output_value(vec[i], i+1, label_c)
    return
    
def set_output_value(value, val_index, label):
    """
    Function to send a value at a specific location of an output vector.  
    
    Parameters
    ----------
    value : float
        Value to be saved.
    val_index : int
        index in the output vector where the value must be saved. This index 
        must be at least 1 and maximum equal to the size specified with the 
        function 'define_output_vector()'.
    label : str
        The name of the output. It must have been defined with the function 
    'define_output_vector()'.
    """
    label_c = str_to_bytes(label)
    libutilities.set_output_value(value, val_index, label_c)
    return
    
def mbs_msg(msg):
    msg = bytes_to_str(msg)
    print(msg, end = '')
    
def mbs_warning(msg):
    msg = bytes_to_str(msg)
    print("MBS-WARNING: "+msg+"\n", end = '')
    
def mbs_error(msg):
    msg = bytes_to_str(msg)
    print("Error encounted in MBsysC core (C-library):")
    print("MBS-ERROR: "+msg+"\n", end = '')

    input("System paused until a key is pressed.\nUnpausing the system may cause kernel restart, erasing all messages!")
    
    raise SystemExit("Some memory may be unset as an error occured in C library.")


mbs_msg_wrapped = mbs_messages_wrap(mbs_msg)
mbs_warning_wrapped = mbs_messages_wrap(mbs_warning)
mbs_error_wrapped = mbs_messages_wrap(mbs_error)

# Redirecting message functions
libutilities.mbs_set_msg_fct_ptr(mbs_msg_wrapped)
libutilities.mbs_set_warning_msg_fct_ptr(mbs_warning_wrapped)
libutilities.mbs_set_error_msg_fct_ptr(mbs_error_wrapped)



