#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
-------------------------------
(c) Universite catholique de Louvain, 2019

Creation : 2019 by LB
Last update : 2019

version MBsysC v1.11.2

-------------------------------
Portable Python interface to MBsysC using Ctypes.

Define the class MbsEquil based on the MbsEquil structure of MBsysC.

"""


import os
import imp
import ctypes

import numpy as np

# importing MbsysPy functions
from ..mbs_utilities import bytes_to_str
from ..mbs_utilities import str_to_bytes

# importing libraries
from .._mbsysc_loader.loadlibs import libmodules

# importing utilities function
from ..mbs_utilities import callback_undefined

# importing wrapping function
from .._mbsysc_loader.callback import user_cons_hJ_wrap
from .._mbsysc_loader.callback import user_cons_jdqd_wrap
from .._mbsysc_loader.callback import user_equil_init_wrap
from .._mbsysc_loader.callback import user_equil_loop_wrap
from .._mbsysc_loader.callback import user_equil_finish_wrap
from .._mbsysc_loader.callback import user_equil_fxe_wrap

from .._mbsysc_loader.callback import mbs_cons_hJ_wrap
from .._mbsysc_loader.callback import mbs_cons_jdqd_wrap
from .._mbsysc_loader.callback import mbs_invdyna_wrap


#==============================================================================
# Global parameter of the current module
#==============================================================================
__DEBUG__ = False
__MODULE_DIR__ = os.path.dirname(os.path.abspath(__file__))


#==============================================================================
# Defining Python MbsEquil class
#==============================================================================

class MbsEquil(object):
    """
    Class of the Equilibrium (static or quasi-static) module.
    
    Attributes
    ----------
    results : MbsResultEquil
        Instance of MbsResultEquil (dedicated to equilibrium module) containing
        the results of the numerical analysis.
    
    Examples
    --------
    >>> mbs_data = MBsysPy.MbsData("../dataR/ExampleProject.mbs")
    >>> mbs_equil = MBsysPy.MbsEquil(mbs_data)
    >>> mbs_equil.set_options(equitol = 1e-4)
    >>> mbs_equil.get_options(equitol)
        0.0001
    >>> results = mbs_equil.run()
    
    """
    def __init__(self, mbs,user_path=None, symbolic_path=None):
        if __DEBUG__ : print( "DEBUG>>  Creating MbsEquil struct for "+mbs.mbs_name+"' MBS.")
        self.mbs_equil_ptr = libmodules.mbs_new_equil(mbs.mbs_data_ptr)
        if __DEBUG__ : print( "DEBUG>>  MbsEquil structure loaded")
        self.mbs = mbs
     
    #Test
        if __DEBUG__ :
            print( "DEBUG>>  MbsEquil created.")
        # Path to user function used by partitionning module
        self.user_path = self.mbs.user_path
        if not (user_path==None):
            project_path  = self.mbs.project_path
            user_path = os.path.join(project_path, user_path)
            # Error handeling
            if not os.path.isdir(user_path):
                print ('The user function directory for direct dynamic module does not exist: "'+ user_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.user_path+'".')
            else:
                self.user_path = user_path
        # Path to user function used by partitionning modue
        self.symbolic_path = self.mbs.symbolic_path
        if not (user_path==None):
            project_path  = self.mbs.project_path
            symbolic_path = os.path.join(project_path, symbolic_path)
            # Error handeling
            if not os.path.isdir(symbolic_path):
                print ('Thesymbolic function directory for direct dynamic module does not exist: "'+ symbolic_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.symbolic_path+'".')
            else:
                self.symbolic_path = symbolic_path
        
        # dictionary containing the pointers to avoid garbage collecting
        self.ptrs_to_user_fcts = dict()
        self.ptrs_to_symb_fcts = dict()
        
        # Storing project function pointer
        self.user_cons_hJ = None
        self.user_cons_jdqd = None
        self.user_equil_init = None
        self.user_equil_loop = None
        self.user_equil_finish = None
        self.mbs_cons_hJ = None
        self.mbs_cons_jdqd = None
        self.mbs_invdyna = None
        
        # exchange of equilibrium variables
        self.exchange_list = []
        
        # addition of equilibrium variables
        self.addition_list = []
        
        # storing result
        self.store_results = True 
        self.results = MbsResultEquil()
    
        # Exposing some memory
        if __DEBUG__ : print( "DEBUG>>  Exposing MbsEquil fields")
        # Constraints
        self._h = self._Jac = self._jdqd = None
        if self.mbs.Ncons>0:
            self._h = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.h, (self.mbs.Ncons+1,))
            self._Jac = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.Jac[0], (self.mbs.Ncons+1,self.mbs.njoint+1))
            self._jdqd = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.jdqd, (self.mbs.Ncons+1,))
        self._huserc = self._Juserc = self._jdqduserc = None
        if self.mbs.Nuserc>0:
            self._huserc = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.huserc, (self.mbs.Nuserc+1,))
            self._Juserc = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.Juserc[0], (self.mbs.Nuserc+1,self.mbs.njoint+1))
            self._jdqduserc = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.jdqduserc, (self.mbs.Nuserc+1,))
        # Fields from equation of motion
        self._M = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.M[0], (self.mbs.njoint+1,self.mbs.njoint+1))
        self._c = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.c, (self.mbs.njoint+1,))
        self._F = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.F, (self.mbs.njoint+1,))
        self._phi = np.ctypeslib.as_array(self.mbs_equil_ptr.contents.aux.contents.phi, (self.mbs.njoint+1,))
        
        # Loading user function
        if self.mbs.opt_load_c < 2:
            if __DEBUG__ : print( "DEBUG>>  Loading user functions")
            self.__load_user_fct__(self.user_path)
        # Loading symbolic function
        if self.mbs.opt_load_c < 1:
            if __DEBUG__ : print( "DEBUG>>  Loading symbolic functions")
            self.__load_symbolic_fct__(self.symbolic_path)
        

    def __str__(self):
        if __DEBUG__ : print( "DEBUG>>  start of __str")
        
        return "MbsEquil instance has nothing to be printed from C library!"
    
    def __load_user_fct__(self, user_path):
        """ 
            Load user function where some args depend on MbsEquil module.
            
            Load the user functions in which some of the arguments are  
            dependent of MbsEquil module instance. The functions will be 
            assigned to the MbsData instance when the 'run' functions is called 
            and unassigned at the end.
            
            The loader user functions are :
             - user_cons_hJ (from user_cons_hJ.py)
             - user_cons_jdqd (from user_cons_jdqd.py)
             - user_equil_init (from user_equil.py)
             - user_equil_loop (from user_equil.py)
             - user_equil_finish (from user_equil.py)
        """
        
        template_path = os.path.join(__MODULE_DIR__,'../templates/user')
        
        
        # cons_hJ
        user_file = "user_cons_hJ.py"
        path = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        module = imp.load_source(user_file[:-3], path)
        self.user_cons_hJ = module.user_cons_hJ
        
        # cons_jdqd
        user_file = "user_cons_jdqd.py"
        path = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        module = imp.load_source(user_file[:-3], path)
        self.user_cons_jdqd = module.user_cons_jdqd
        
        # user_equil
        user_file = "user_equil.py"
        path      = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            print ("file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        else:
            if __DEBUG__ : print( "DEBUG>>  loading file '"+user_file+"' in folder '"+os.path.dirname(path))
        module = imp.load_source(user_file[:-3], path)
        self.user_equil_init   = module.user_equil_init
        self.user_equil_loop   = module.user_equil_loop
        self.user_equil_finish = module.user_equil_finish
        try: 
            self.user_equil_fxe = module.user_equil_fxe
        except AttributeError:
            path = os.path.abspath(os.path.join(template_path, user_file))
            module = imp.load_source(user_file[:-3], path)
            self.user_equil_fxe = module.user_equil_fxe
        return
        
    def __load_symbolic_fct__(self, symbolic_path):
        """ 
            Load symbolic function where some args depend on MbsEquil module.
            
            Load the symb functions in which some of the arguments are  
            dependent of MbsEquil module instance. The functions will be 
            assigned to the MbsData instance when the 'run' functions is called 
            and unassigned at the end.
            
            The loader user functions are :
             - cons_hJ (from mbs_cons_hJ_MBSNAME.py)
             - cons_jdqd (from mbs_cons_jdqd_MBSNAME.py)
             - dirdyna (from mbs_dirdyna_MBSNAME.py)
             - invdyna (from mbs_invdyna_MBSNAME.py)
        """
        mbs_name = self.mbs.mbs_name
        template_path = os.path.join(__MODULE_DIR__,'../templates/symbolic')
        # mbs_cons_hJ
        symb_file = "mbs_cons_hJ_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_cons_hJ_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_cons_hJ = module.cons_hJ
        
        
        # mbs_cons_jdqd
        symb_file = "mbs_cons_jdqd_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_cons_jdqd_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_cons_jdqd = module.cons_jdqd
        
        # invdyna
        symb_file = "mbs_invdyna_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_invdyna_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_invdyna = module.invdyna
        return
    
    def __assign_user_fct__(self):
        """ 
            Assign and wrap python user function where some args depend 
            on MbsEquil module. Store the functions in the MbsData instance 
            and assign the pointer of functions in the C structure
        """
        # cons_hJ
        self.ptrs_to_user_fcts["user_cons_hJ"] = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.user_cons_hJ(self._huserc, self._Juserc, self.mbs, tsim))
        self.mbs.user_cons_hJ = self.user_cons_hJ
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = self.ptrs_to_user_fcts["user_cons_hJ"]
        
        # cons_jdqd
        self.ptrs_to_user_fcts["user_cons_jdqd"] = user_cons_jdqd_wrap(lambda jdqd,mbs,tsim : self.user_cons_jdqd(self._jdqduserc, self.mbs))
        self.mbs.user_cons_jdqd = self.user_cons_jdqd
        self.mbs.mbs_data_ptr.contents.user_cons_jdqd = self.ptrs_to_user_fcts["user_cons_jdqd"]

        # user_equil
        self.ptrs_to_user_fcts["user_equil_init"] = user_equil_init_wrap(lambda mbs, eq : self.mbs.user_equil_init(self.mbs, self))
        self.ptrs_to_user_fcts["user_equil_loop"] = user_equil_loop_wrap(lambda mbs, eq : self.mbs.user_equil_loop(self.mbs, self))
        self.ptrs_to_user_fcts["user_equil_finish"] = user_equil_finish_wrap(lambda mbs, eq : self.mbs.user_equil_finish(self.mbs, self))
        self.ptrs_to_symb_fcts["user_equil_fxe"] = user_equil_fxe_wrap(lambda mbs, f : self.__callback_user_equil_fxe(self.user_equil_fxe, self.mbs, f))
        self.mbs.mbs_data_ptr.contents.user_equil_fxe = self.ptrs_to_symb_fcts["user_equil_fxe"]
        self.mbs.user_equil_fxe = self.user_equil_fxe
        self.mbs.user_equil_init = self.user_equil_init
        self.mbs.mbs_data_ptr.contents.user_equil_init = self.ptrs_to_user_fcts["user_equil_init"]
        self.mbs.user_equil_loop = self.user_equil_loop
        self.mbs.mbs_data_ptr.contents.user_equil_loop = self.ptrs_to_user_fcts["user_equil_loop"]
        self.mbs.user_equil_finish = self.user_equil_finish
        self.mbs.mbs_data_ptr.contents.user_equil_finish = self.ptrs_to_user_fcts["user_equil_finish"]
        return
        
    def __assign_symbolic_fct__(self):
        """ 
            Assign and wrap python symbolic function where some args depend 
            on MbsEquil module. Store the functions in the MbsData instance 
            and assign the pointer of functions in the C structure
        """
        
        # mbs_cons_hJ
        self.ptrs_to_symb_fcts["mbs_cons_hJ"] = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.mbs_cons_hJ(self._h,self._Jac, self.mbs))
        self.mbs.mbs_cons_hJ = self.mbs_cons_hJ
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = self.ptrs_to_symb_fcts["mbs_cons_hJ"]
        
        # mbs_cons_jdqd
        self.ptrs_to_symb_fcts["mbs_cons_jdqd"] = mbs_cons_jdqd_wrap(lambda jdqd,mbs,tsim : self.mbs_cons_jdqd(self._jdqd, self.mbs))
        self.mbs.mbs_cons_jdqd = self.mbs_cons_jdqd
        self.mbs.mbs_data_ptr.contents.mbs_cons_jdqd = self.ptrs_to_symb_fcts["mbs_cons_jdqd"]
        
        # invdyna
        self.ptrs_to_symb_fcts["mbs_invdyna"] = mbs_invdyna_wrap(lambda phi,mbs,tsim : self.mbs_invdyna(self._phi, self.mbs, tsim))
        self.mbs.mbs_invdyna = self.mbs_invdyna
        self.mbs.mbs_data_ptr.contents.mbs_invdyna = self.ptrs_to_symb_fcts["mbs_invdyna"]
         
        return
    
     # Callback function for function with advanced arguments
    def __callback_mbs_cons_hJ(self, fun, h, Jac):
        if __DEBUG__ : print( "DEBUG>>  callback_mbs_cons_hJ")
        __h = np.ctypeslib.as_array(h, (self.mbs.Ncons+1,))
        __Jac = np.ctypeslib.as_array(Jac[0], (self.mbs.Ncons+1,self.mbs.njoint+1))
        fun(__h,__Jac,self.mbs)
    
    def __callback_user_equil_fxe(self, fun, mbs, f):
        __f = np.ctypeslib.as_array(f, (self.mbs_equil_ptr.contents.nx+1,))
        fun(mbs,__f)
        
    def __unassign_user_fct__(self):
        """ 
            Unassign user function where some args depend on MbsEquil module.
        """
        
        self.ptrs_to_user_fcts = {}
        
        self.mbs.user_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("user_cons_hJ"))
        self.mbs.user_cons_jdqd = None
        self.mbs.mbs_data_ptr.contents.user_cons_jdqd = user_cons_jdqd_wrap(lambda h,Jac,mbs,tsim : callback_undefined("user_cons_jdqd"))
        self.mbs.user_equil_init = None
        self.mbs.mbs_data_ptr.contents.user_equil_init = user_equil_init_wrap(lambda mbs, eq : callback_undefined("user_equil_init"))
        self.mbs.user_equil_loop = None
        self.mbs.mbs_data_ptr.contents.user_equil_loop = user_equil_loop_wrap(lambda mbs, eq : callback_undefined("user_equil_loop"))
        self.mbs.user_equil_finish = None
        self.mbs.mbs_data_ptr.contents.user_equil_finish = user_equil_finish_wrap(lambda mbs, eq : callback_undefined("user_equil_finish"))
        self.mbs.user_equil_fxe = None
        self.mbs.mbs_data_ptr.contents.user_equil_fxe = user_equil_fxe_wrap(lambda mbs, f : callback_undefined("user_equil_fxe"))
        
    def __unassign_symbolic_fct__(self):
        
        """ 
            Unassign symbolic function where some args depend on MbsEquil module.
        """
        
        self.ptrs_to_symb_fcts = {}
        
        self.mbs.mbs_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("mbs_cons_hJ"))
        self.mbs.mbs_cons_jdqd = None
        self.mbs.mbs_data_ptr.contents.mbs_cons_jdqd = mbs_cons_jdqd_wrap(lambda jdqd,mbs,tsim : callback_undefined("mbs_cons_jdqd"))
        self.mbs.mbs_invdyna = None
        self.mbs.mbs_data_ptr.contents.mbs_invdyna = mbs_invdyna_wrap(lambda Q,mbs,tsim : callback_undefined("mbs_invdyna"))
        
    def __del__(self):
        
        libmodules.mbs_delete_equil(self.mbs_equil_ptr, self.mbs.mbs_data_ptr)
        if __DEBUG__ : print("DEBUG>>  MbsEquil pointer deleted")
        
    def run(self):
        """
            Run an equil
            
            Options can be setted with set_options 
            Options can be retrieved with get_options
        
        """
        # Assing required user functions
        if self.mbs.opt_load_c < 2:
            self.__assign_user_fct__()
            self.mbs.__assign_user_fct__()
        
            self._compute_exchange()
            self._compute_extra_variable()
        
        # Assing required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__assign_symbolic_fct__()
            self.mbs.__assign_symb_fct__()
        
        libmodules.mbs_run_equil(self.mbs_equil_ptr, self.mbs.mbs_data_ptr)
        
        if self.store_results:
            self.load_results()
            
        # Anassign user functions
        if self.mbs.opt_load_c < 2:
            self.__unassign_user_fct__()
            self.mbs.__unassign_user_fct__()
        
        # Anassing required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__unassign_symbolic_fct__()
            self.mbs.__unassign_symb_fct__()
        
        return self.results
        
        print(">>EQUIL>> MbsData structure contains the results of the equilibrium")
        
    def load_results(self):
        
        mode = self.get_options("mode")
        if mode == 1:
            self.results.type = "static"
        elif mode == 2:
            self.results.type = "quasistatic"
        elif mode == 3:
            self.results.type = "dynamic"
        else:
            self.results.type = "not an existing mode"
            
        self.results.nquch = self.get_options("nquch")
        self.results.nxe = self.get_options("nxe")
        self.results.nx = self.nx 
        
        self.results.x = np.copy(np.ctypeslib.as_array(self.mbs_equil_ptr.contents.x, (self.nx,)))
        for i, val in enumerate(self.results.x):
            self.results.x_name.append("q" + str(i+1))
        if self.mbs.Nux and self.get_options("compute_uxd"):
            for i in range(self.mbs.Nux):
               self.results.x_name[self.nx-self.mbs.Nux+i] = "ux" + str(i+1)
        if self.results.nxe:
            for i, val in enumerate(self.addition_list):
                self.results.x_name[self.nx-self.results.nxe+i] = self.addition_list[i][1]
        if self.results.nquch:
            for i, val in enumerate(self.exchange_list):
                self.results.x_name[self.exchange_list[i][1]-1] = self.exchange_list[i][2]
        if self.success == 1:
            self.results.success = "PASSED"
        elif self.success == 0:
            self.results.success = "NONE"
        elif self.success == -1:
            self.results.success = "FAILED"
            
    def add_exchange(self, ptr, index, replaced_var_id, ptr_name = "??"):
        """ Add a new equilibrium variable that replaces the specified 
            default one.
            
            Exchanged variables are stored and will be set in the memory of the 
            external library when 'MbsEquil::run()' is called. Between two run()
            the equilibrium can be modified thanks to replace(). 
            However as the number of equilibrium variables is locked in the
            external memory after the first call of run(), this function can not
            be called between two run().
            
            Parameters
            ----------
            
            ptr : ndarray
                Array containing the variable to be exchanged
            index : int
                Index in ptr of the variable to be exchanged
            replaced_var_id : int
                Index of the default variable that will be replaced
            ptr_name : str
                (optionnal) Name of the variable to be exchanged
                                
        """
        if self.get_options("nquch"):
            print("As the number of equilibrium variables is locked in the external memory after the first call of run(), this function can not be called between two run()")
        if index > np.size(ptr)-1:
            print(">>EQUIL>> variable exchange : Index out of bound")
            return
        address = ptr.ctypes.data + ctypes.sizeof(ctypes.c_double)*index
        if ptr.ctypes.data == self.mbs.q.ctypes.data:
            print(">>EQUIL>> variable exchange : Warning ! Generalized coordinates have to be used with caution as exchange variables")
        
        if len(self.exchange_list) == 0:
            self.exchange_list = [[address, replaced_var_id, ptr_name]]
        else:
            for element in self.exchange_list:
                if element[0] == address:
                    print(">>EQUIL>> variable exchange : Exchange variable is already an equilibrium variable")
                else:
                    self.exchange_list.append([address, replaced_var_id, ptr_name])
                    return
                    
    def add_extra_variable(self, ptr, index, ptr_name = "??"):
        """ Add a new equilibrium variable
            
            Extra variables are stored and will be set in the memory of the 
            external library when 'MbsEquil::run()' is called. Between two run()
            the equilibrium can be modified thanks to replace(). 
            However as the number of equilibrium variables is locked in the
            external memory after the first call of run(), this function can not
            be called between two run().
            
            Parameters
            ----------
            
            ptr : ndarray
                Array containing the extra variable to add to the equilibrium variables
            index : int
                Index in ptr of the extra variable to add to the equilibrium variables
            ptr_name : str 
                (optionnal) Name of the extra variable
            
        """
        if self.get_options("nxe"):
            print("As the number of equilibrium variables is locked in the external memory after the first call of run(), this function can not be called between two run()")
        if index > np.size(ptr)-1:
            print(">>EQUIL>> variable exchange : Index out of bound")
            return
        address = ptr.ctypes.data + ctypes.sizeof(ctypes.c_double)*index
        if ptr.ctypes.data == self.mbs.q.ctypes.data:
            print(">>EQUIL>> extra exchange : Warning ! Generalized coordinates have to be used with caution as extra variables")
        
        if len(self.addition_list) == 0:
            self.addition_list = [[address, ptr_name]]
        else:
            for element in self.addition_list:
                if element[0] == address:
                    print(">>EQUIL>> variable exchange : Extra variable has already been added")
                else:
                    self.addition_list.append([address, ptr_name])
                    return
         
    def add_extra_variable_um(self, user_model, param):
        """ Add a user model as new equilibrium variable 
            
            Extra variables are stored and will be set in the memory of the 
            external library when 'MbsEquil::run()' is called. Between two run()
            the equilibrium can be modified thanks to replace(). 
            However as the number of equilibrium variables is locked in the
            external memory after the first call of run(), this function can not
            be called between two run().
            
            Parameters
            ----------
            
            user_model : str
                Name of the user model to add to the equilibrium variables
            param : str
                Parameter of the user model to add to the equilibrium variables
            
        """
        if self.get_options("nxe"):
            print("As the number of equilibrium variables is locked in the external memory after the first call of run(), this function can not be called between two run()")
        if self.mbs.user_model[user_model]._type[param] != 1 and self.mbs.user_model[user_model]._type[param] != 7:
            print(">>EQUIL>> variable exchange : This function does not handle this type of user model (only float and int)")
            return
        address = self.mbs.user_model[user_model].__getarray__(param)[0].ctypes.data
        if __DEBUG__ : print( "DEBUG>>  add_extra_variable_um, address: ", address)
        if len(self.addition_list) == 0:
            self.addition_list = [[address, user_model+ "." + param]]
        else:
            for element in self.addtion_list:
                if element[0] == address:
                    print(">>EQUIL>> variable exchange : Extra variable has already been added")
                else:
                    self.addition_list.append([address, user_model+ "." + param])
                    return
            
    def add_exchange_um(self, user_model, param, replaced_var_id):
        """ Add a user model as a new equilibrium variable that replaces 
            the specified default one
            
            Exchanged variables are stored and will be set in the memory of the 
            external library when 'MbsEquil::run()' is called. Between two run()
            the equilibrium can be modified thanks to replace(). 
            However as the number of equilibrium variables is locked in the
            external memory after the first call of run(), this function can not
            be called between two run().
            
            Parameters
            ----------
            
            user_model : str
                Name of the user model to be exchanged
            param : str
                Parameter of the user model to be exchanged
            replaced_var_id : int
                Index of the default variable that will be replaced
            
        """
        if __DEBUG__ : print( "DEBUG>>  add_exchange_um, parameters: ", user_model, param, replaced_var_id)
        if self.get_options("nquch"):
            print("As the number of equilibrium variables is locked in the external memory after the first call of run(), this function can not be called between two run()")
            return
        if self.mbs.user_model[user_model]._type[param] != 1 and self.mbs.user_model[user_model]._type[param] != 7:
            print(">>EQUIL>> variable exchange : This function does not handle this type of user model (only float and int)")
            return
        address = self.mbs.user_model[user_model].__getarray__(param)[0].ctypes.data
        if __DEBUG__ : print( "DEBUG>>  add_exchange_um, address: ", address)
        
        if len(self.exchange_list) == 0:
            if __DEBUG__ : print( "DEBUG>>  add_exchange_um, first exchanged variable") 
            self.exchange_list = [[address, replaced_var_id, user_model+ "." + param]]
        else:
            for element in self.exchange_list:
                if __DEBUG__ : print( "DEBUG>>  add_exchange_um, iterating exchanged variables:", element) 
                if element[0] == address:
                    print(">>EQUIL>> variable exchange : Exchange variable is already an equilibrium variable")
                    exit()
                else:
                    if __DEBUG__ : print( "DEBUG>>  add_exchange_um, vatiable added") 
                    self.exchange_list.append([address, replaced_var_id, user_model+ "." + param])
                    return
        
        
    def _compute_exchange(self):
        """
           Set exchanged variables in the memory of the external library when 
           'MbsEquil::run()' is called
        """
        if self.get_options("nquch") == 0 and len(self.exchange_list) > 0:
            self.mbs_equil_ptr.contents.options.contents.nquch = len(self.exchange_list)
            libmodules.mbs_equil_exchange(self.mbs_equil_ptr.contents.options)
        if self.get_options("nquch") == len(self.exchange_list) and len(self.exchange_list) > 0:
            for i, element in enumerate(self.exchange_list):
                libmodules.mbs_equil_set_variable(self.mbs_equil_ptr.contents.options, element[0],i+1, element[1])
        elif self.get_options("nquch") != len(self.exchange_list):
            print(">>EQUIL>> variable exchange : nquch (" +str(self.get_options("nquch"))+ ") different from the length of the exchange list ("+ str(len(self.exchange_list))+")")
       
    def _compute_extra_variable(self):
        """
           Set extra variables in the memory of the external library when 
           'MbsEquil::run()' is called
        """
        if self.get_options("nxe") == 0 and len(self.addition_list) > 0:
            self.mbs_equil_ptr.contents.options.contents.nxe = len(self.addition_list)
            libmodules.mbs_equil_addition(self.mbs_equil_ptr.contents.options)
        if self.get_options("nxe") == len(self.addition_list) and len(self.addition_list) > 0:
            for i, element in enumerate(self.addition_list):
                libmodules.mbs_equil_add_variable(self.mbs_equil_ptr.contents.options, element[0],i+1)
        elif self.get_options("nxe") != len(self.addition_list):
            print(">>EQUIL>> extra variable : nxe (" +str(self.get_options("nxe"))+ ") different from the length of the extra variable list ("+ str(len(self.addition_list))+")")
        
    def replace_exchange(self, ptr, index, id_exchanged, ptr_name = "??"):
        """ Replace an existing exchanged variable with another one
        
            This function can be called between two 'MbsEquil::run()' as it
            does not change the number of equilibrium variables
            
            
            Parameters
            ----------
            
            ptr : ndarray
                Array containing the variable to be exchanged
            index : int
                Index in ptr of the variable to be exchanged
            id_exchanged : int
                Index in the list of exchanged variable of the variable to be 
                replaced
            ptr_name : str 
                (optionnal) Name of the variable to be exchanged
            
        """
        if ptr.ctypes.data == self.mbs.q.ctypes.data:
            print(">>EQUIL>> variable exchange : Warning   generalized coordinates have to be used with caution as exchange variables")
        
        if id_exchanged > len(self.exchange_list) or id_exchanged <=0 :
            print(">>EQUIL>> variable exchange : There is no exchange variable at this index :" + str(id_exchanged))
        else:
            self.exchange_list[id_exchanged-1][0] = ptr.ctypes.data + ctypes.sizeof(ctypes.c_double)*index
            self.exchange_list[id_exchanged-1][2] = ptr_name
            
    def replace_extra_variable(self, ptr, index, id_extra, ptr_name = "??"):
        """ Replace an existing extra variable with another one
        
            This function can be called between two 'MbsEquil::run()' as it
            does not change the number of equilibrium variables
            
            
            Parameters
            ----------
            
            ptr : ndarray
                Array containing the extra variable
            index : int
                Index in ptr of the extra variable
            id_exchanged : int
                Index in the list of extra variable of the variable to be 
                replaced
            
        """
        if ptr.ctypes.data == self.mbs.q.ctypes.data:
            print(">>EQUIL>> extra exchange : Warning   generalized coordinates have to be used with caution as extra variables")
        
        if id_extra > len(self.addition_list) or id_extra <=0 :
            print(">>EQUIL>> extra exchange : There is no extra variable at this index :" + str(id_extra))
        else:
            self.addition_list[id_extra-1][0] = ptr.ctypes.data + ctypes.sizeof(ctypes.c_double)*index
            self.addition_list[id_extra-1][1] = ptr_name
        
    def replace_extra_variable_um(self, user_model, param, id_extra):
        """ Replace an existing extra variable with a user model
        
            This function can be called between two 'MbsEquil::run()' as it
            does not change the number of equilibrium variables
            
            
            Parameters
            ----------
            
            user_model : str
                Name of the user model to add to the equilibrium variables
            param : str
                Parameter of the user model to add to the equilibrium variables
            id_exchanged : int
                Index in the list of extra variable of the variable to be 
                replaced
            
        """
        if self.mbs.user_model[user_model]._type[param] != 1 and self.mbs.user_model[user_model]._type[param] != 7:
            print(">>EQUIL>> extra exchange : This function does not handle this type of user model (only float and int)")
            return
        if id_extra > len(self.addition_list) or id_extra <=0 :
            print(">>EQUIL>> extra exchange : There is no extra variable at this index :" + str(id_extra))
        else:
            self.addition_list[id_extra-1][0] = self.mbs.user_model[user_model].__getarray__(param)[0].ctypes.data     
            self.addition_list[id_extra-1][1] = user_model+ "." + param
    
    def replace_exchange_um(self, user_model, param, id_exchanged):
        """ Replace an existing exchanged variable with a user model
        
            This function can be called between two 'MbsEquil::run()' as it
            does not change the number of equilibrium variables
            
            
            Parameters
            ----------
            
            user_model : str
                Name of the user model to be exchanged
            param : str
                Parameter of the user model to be exchanged
            id_exchanged : int
                Index in the list of exchanged variable of the variable to be 
                replaced
            
        """
        if self.mbs.user_model[user_model]._type[param] != 1 and self.mbs.user_model[user_model]._type[param] != 7:
            print(">>EQUIL>> variable exchange : This function does not handle this type of user model (only float and int)")
            return
        if id_exchanged > len(self.exchange_list) or id_exchanged <=0 :
            print(">>EQUIL>> variable exchange : There is no exchange variable at this index :" + str(id_exchanged))
        else:
            self.exchange_list[id_exchanged-1][0] = self.mbs.user_model[user_model].__getarray__(param)[0].ctypes.data 
            self.exchange_list[id_exchanged-1][2] = user_model+ "." + param
        
    def print_equil(self):
        
        libmodules.mbs_print_equil(self.mbs_equil_ptr)
        

    def set_options(self, **kwargs):
        """
        Set the specified options for Equil module
        
        Parameters
        ----------
        
        method : int
            fsolvepk = 1 / optim = 2
            default = 1
        equitol : float
            equilibrium acceptation tolerance
            default = 1e-6
        verbose : int
            no = 0, yes = 1
            default = 1
        save2file : int
            1: results saved, 0: not saved    
            default = 1
        compute_uxd : int
            no = 0, yes = 1
            flag to compute the extra constitutive differential equations as equil equations
            default = 1
        grad_lpk : int
            options to compute the gradient through the lpk parabolic fitting method.
            If not activated, the gradient is computed by small deviation with options devJac
            default = 0
        lpk_itermax : int 
            options for the lpk gradient
            default = 10
        lpk_relincr : float
            options for the lpk gradient
            default = 1e-2
        lpk_absincr : float
            options for the lpk gradient
            default = 1e-3
        lpk_equitol : float
            options for the lpk gradient
            default = 1e-6
        lpk_lintol : float
            options for the lpk gradient
            default = 1e-3
        resfilename : char
            The keyword used for determining the name of result files
            default : equil
        respath : char
            Path in which result file are saved.    
            default : resultsR folder
        animpath : char
            Path in which animation file are saved.
            default : animationR folder
        mode : int 
            1=static / 2=quasistatic / 3=dynamic    
            default = 0
        relax : int
            relaxation in pk algorithm
            no = 0, yes = 1    
            default = 0
        relaxcoeff : float
            relaxation coefficient
            default = 0.9
        relaxIterMax : float
            maximum number of relaxation iterations
            default = 10
        soft : int 
            softing in pk algorithm
            no = 0, yes = 1
            default = 0
        senstol : float
            tolerance for testing a variable non-sensitivity    
            default = 1e-6
        devjac : float 
            deviate state for Gradient calculation
            default = 1e-6
        itermax : int 
            maximum number of iteration for solvepk    
            default = 30
        visualize : int
            no = 0, yes = 1
            default = 0
        clearmbsglobal : int
            inout = 1, out = 2, none = 3, all = 4
            default = 1 
        saveperiod : int 
            The number of iteration between two buffer records
            default = 1 (every iteration are recorded)
        save_anim : int
            1 to save the anim file
            0 otherwise (only valid if 'save2file' is set to 1)
        framerate : int
            number of frame per second for the .anim file
        buffersize : int
            The number of time step that can be recorded in the buffer
            default = -1
        max_save_user : int 
            The maximal number of user variables saved
            default = 12
        """
        for key, value in kwargs.items():
            if key == "method":
                self.mbs_equil_ptr.contents.options.contents.methods = value
            elif key == "equitol":
                self.mbs_equil_ptr.contents.options.contents.equitol = value
            elif key == "verbose":
                self.mbs_equil_ptr.contents.options.contents.verbose = value
            elif key == "save2file":
                self.mbs_equil_ptr.contents.options.contents.save2file = value
            elif key == "compute_uxd":
                self.mbs_equil_ptr.contents.options.contents.compute_uxd = value
            elif key == "grad_lpk":
                self.mbs_equil_ptr.contents.options.contents.grad_lpk = value
            elif key == "lpk_itermax":
                self.mbs_equil_ptr.contents.options.contents.lpk_itermax = value
            elif key == "lpk_relincr":
                self.mbs_equil_ptr.contents.options.contents.lpk_relincr = value
            elif key == "lpk_absincr":
                self.mbs_equil_ptr.contents.options.contents.lpk_absincr = value
            elif key == "lpk_equitol":
                self.mbs_equil_ptr.contents.options.contents.lpk_equitol = value
            elif key == "lpk_lintol":
                self.mbs_equil_ptr.contents.options.contents.lpk_lintol = value
            elif key == "resfilename":
                self.mbs_equil_ptr.contents.options.contents.resfilename = str_to_bytes(value)
            elif key == "respath":
                self.mbs_equil_ptr.contents.options.contents.respath = str_to_bytes(value)
            elif key == "animpath":
                self.mbs_equil_ptr.contents.options.contents.animpath = str_to_bytes(value)
            elif key == "mode":
                self.mbs_equil_ptr.contents.options.contents.mode = value
            elif key == "relax":
                self.mbs_equil_ptr.contents.options.contents.relax = value
            elif key == "relaxcoeff":
                self.mbs_equil_ptr.contents.options.contents.relaxcoeff = value
            elif key == "relaxIterMax":
                self.mbs_equil_ptr.contents.options.contents.relaxIterMax = value
            elif key == "soft":
                self.mbs_equil_ptr.contents.options.contents.soft = value
            elif key == "senstol":
                self.mbs_equil_ptr.contents.options.contents.senstol = value
            elif key == "devjac":
                self.mbs_equil_ptr.contents.options.contents.devjac = value
            elif key == "itermax":
                self.mbs_equil_ptr.contents.options.contents.itermax = value
            elif key == "visualize":
                self.mbs_equil_ptr.contents.options.contents.visualize = value
            elif key == "clearmbsglobal":
                self.mbs_equil_ptr.contents.options.contents.clearmbsglobal = value
            elif key == "save_anim":
                self.mbs_equil_ptr.contents.options.contents.save_anim = value
            elif key == "framerate":
                self.mbs_equil_ptr.contents.options.contents.framerate = value
            elif key == "saveperiod":
                self.mbs_equil_ptr.contents.options.contents.saveperiod= value
            elif key == "max_save_user":
                self.mbs_equil_ptr.contents.options.contents.max_save_user = value
            elif key == "buffersize":
                self.mbs_equil_ptr.contents.options.contents.buffersize = value
            elif key == "store_results":
                self.store_results = value
            else:
                print(">>EQUIL>  The option "+key+ " is not defined in this module")
                
                
    def get_options(self, *args):
        """
        Get the specified options for Equil module
        
        Parameters
        ----------
        
        The different options specifed in the documentation of set_options
        
        Returns
        -------
    
        The value of the different options specifed in the documentation of set_options
        
        
        """
        options = []
        for key in args:
            if key == "method":
                options.append(self.mbs_equil_ptr.contents.options.contents.methods)
            elif key == "equitol":
                options.append(self.mbs_equil_ptr.contents.options.contents.equitol)
            elif key == "verbose":
                options.append(self.mbs_equil_ptr.contents.options.contents.verbose)
            elif key == "save2file":
                options.append(self.mbs_equil_ptr.contents.options.contents.save2file)
            elif key == "compute_uxd":
                options.append(self.mbs_equil_ptr.contents.options.contents.compute_uxd)
            elif key == "grad_lpk":
                options.append(self.mbs_equil_ptr.contents.options.contents.grad_lpk)
            elif key == "lpk_itermax":
                options.append(self.mbs_equil_ptr.contents.options.contents.lpk_itermax)
            elif key == "lpk_relincr":
                options.append(self.mbs_equil_ptr.contents.options.contents.lpk_relincr)
            elif key == "lpk_absincr":
                options.append(self.mbs_equil_ptr.contents.options.contents.lpk_absincr)
            elif key == "lpk_equitol":
                options.append(self.mbs_equil_ptr.contents.options.contents.lpk_equitol)
            elif key == "lpk_lintol":
                options.append(self.mbs_equil_ptr.contents.options.contents.lpk_lintol)
            elif key == "resfilename":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_equil_ptr.contents.options.contents.resfilename)))
            elif key == "respath":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_equil_ptr.contents.options.contents.respath)))
            elif key == "animpath":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_equil_ptr.contents.options.contents.animpath)))
            elif key == "mode":
                options.append(self.mbs_equil_ptr.contents.options.contents.mode)
            elif key == "relax":
                options.append(self.mbs_equil_ptr.contents.options.contents.relax)
            elif key == "relaxcoeff":
                options.append(self.mbs_equil_ptr.contents.options.contents.relaxcoeff)
            elif key == "relaxIterMax":
                options.append(self.mbs_equil_ptr.contents.options.contents.relaxIterMax)
            elif key == "soft":
                options.append(self.mbs_equil_ptr.contents.options.contents.soft)
            elif key == "senstol":
                options.append(self.mbs_equil_ptr.contents.options.contents.senstol)
            elif key == "devjac":
                options.append(self.mbs_equil_ptr.contents.options.contents.devjac)
            elif key == "itermax":
                options.append(self.mbs_equil_ptr.contents.options.contents.itermax)
            elif key == "visualize":
                options.append(self.mbs_equil_ptr.contents.options.contents.visualize)
            elif key == "clearmbsglobal":
                options.append(self.mbs_equil_ptr.contents.options.contents.clearmbsglobal)
            elif key == "nquch":
                options.append(self.mbs_equil_ptr.contents.options.contents.nquch)
            elif key == "nxe":
                options.append(self.mbs_equil_ptr.contents.options.contents.nxe)
            elif key == "save_anim":
                options.append(self.mbs_equil_ptr.contents.options.contents.save_anim)
            elif key == "framerate":
                options.append(self.mbs_equil_ptr.contents.options.contents.framerate)
            elif key == "saveperiod":
                options.append(self.mbs_equil_ptr.contents.options.contents.saveperiod)
            elif key == "max_save_user":
                options.append(self.mbs_equil_ptr.contents.options.contents.max_save_user)
            elif key == "buffersize":
                options.append(self.mbs_equil_ptr.contents.options.contents.buffersize)
            elif key == "store_results":
                options.append(self.store_results)
            else:
                print(">>EQUIL>>  The option "+key+ " is not defined in this module")
                
        if len(options) == 0:
            return
        if len(options) == 1:
            return options[0]
            
        return tuple(options)
            

    #==========================================================================
    # Defining properties
    #==========================================================================
    @property
    def mbs_filename(self):
        return self.mbs.mbs_filename
    
    @property
    def project_path(self):
        return self.mbs.project_path
    
    @property
    def mbs_name(self):
        return self.mbs.mbs_name
    
    @property 
    def nx(self):
        return self.mbs_equil_ptr.contents.nx
    
    @property
    def success(self):
        return self.mbs_equil_ptr.contents.success
    
class MbsResultEquil(object):
    
    def __init__(self):
        
        self.type = None 
        self.nquch = None 
        self.nxe = None
        self.nx = None 
        self.x = None
        self.x_name = []
        self.success = None
        