#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep  2 14:54:43 2019

@author: LB
"""

"""
-------------------------------
(c) Universite catholique de Louvain, 2019

Creation : 2019 by O. Lantsoght
Last update : 2019

version MBsysC v1.11.2

-------------------------------
Portable Python interface to MBsysC using Ctypes.

Define the class MbsInvdyn based on the MbsInvdyn structure of MBsysC.
This class has the functions required to manipulate the direct dynamic module.
This include setting the options, running an (or multiple) analysis and freeing
the memory.
"""

import os
import imp
import ctypes

import numpy as np

# importing MbsysPy functions
from ..mbs_utilities import bytes_to_str
from ..mbs_utilities import str_to_bytes

# importing MbsysPy classes
from .mbs_dirdyn import MbsResult

# importing libraries
from .._mbsysc_loader.loadlibs import libmodules

# importing utilities function
from ..mbs_utilities import callback_undefined

# importing wrapping function
from .._mbsysc_loader.callback import user_cons_hJ_wrap
from .._mbsysc_loader.callback import user_cons_jdqd_wrap

from .._mbsysc_loader.callback import mbs_cons_hJ_wrap
from .._mbsysc_loader.callback import mbs_cons_jdqd_wrap
from .._mbsysc_loader.callback import mbs_invdyna_wrap
from .._mbsysc_loader.callback import mbs_dirdyna_wrap


#==============================================================================
# Global parameter of the current module
#==============================================================================
__DEBUG__ = False
__MODULE_DIR__ = os.path.dirname(os.path.abspath(__file__))


#==============================================================================
# Defining Python MbsInvdyn class
#==============================================================================

class MbsInvdyn(object):
    
    
    def __init__(self, mbs, user_path=None, symbolic_path=None):
        
        self.module_name = type(self).__name__
        
        if __DEBUG__ : print( "DEBUG>>  Creating " +self.module_name + " struct for "+mbs.mbs_name+"' MBS.")
        self.mbs_invdyn_ptr = libmodules.mbs_new_invdyn(mbs.mbs_data_ptr)
        if __DEBUG__ : print( "DEBUG>>  " +self.module_name + " structure loaded")
        self.mbs = mbs
        
        
        self.user_path = self.mbs.user_path
        if not (user_path==None):
            project_path  = bytes_to_str(self.mbs.project_path)
            user_path = os.path.join(project_path, user_path)
            # Error handeling
            if not os.path.isdir(user_path):
                print ('The user function directory for direct dynamic module does not exist: "'+ user_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.user_path+'".')
            else:
                self.user_path = user_path
        # Path to user function used by partitionning modue
        self.symbolic_path = self.mbs.symbolic_path
        if not (user_path==None):
            project_path  = bytes_to_str(self.mbs.project_path)
            symbolic_path = os.path.join(project_path, symbolic_path)
            # Error handeling
            if not os.path.isdir(symbolic_path):
                print ('Thesymbolic function directory for direct dynamic module does not exist: "'+ symbolic_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.symbolic_path+'".')
            else:
                self.symbolic_path = symbolic_path
        
        # dictionary containing the pointers to avoid garbage collecting
        self.ptrs_to_user_fcts = dict()
        self.ptrs_to_symb_fcts = dict()
        
        # Storing project function pointer
        self.user_cons_hJ = None
        self.user_cons_jdqd = None
        self.mbs_cons_hJ = None
        self.mbs_cons_jdqd = None
        self.mbs_invdyna = None
        
        # Storing Results
        self.results = MbsResult(self.mbs)
        self.store_results = True
        
        # Exposing some memory
        if __DEBUG__ : print( "DEBUG>>  Exposing " +self.module_name + " fields")
        # Constraints
        self._h = self._Jac = self._jdqd = None
        if self.mbs.Ncons>0:
            self._h = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.h, (self.mbs.Ncons+1,))
            self._Jac = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.Jac[0], (self.mbs.Ncons+1,self.mbs.njoint+1))
            self._jdqd = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.jdqd, (self.mbs.Ncons+1,))
        self._huserc = self._Juserc = self._jdqduserc = None
        if self.mbs.Nuserc>0:
            self._huserc = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.huserc, (self.mbs.Nuserc+1,))
            self._Juserc = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.Juserc[0], (self.mbs.Nuserc+1,self.mbs.njoint+1))
            self._jdqduserc = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.jdqduserc, (self.mbs.Nuserc+1,))
        # Fields from equation of motion
        self._M = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.M[0], (self.mbs.njoint+1,self.mbs.njoint+1))
        self._c = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.c, (self.mbs.njoint+1,))
        self._F = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.F, (self.mbs.njoint+1,))
        self._phi = np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.mbs_aux.contents.phi, (self.mbs.njoint+1,))
        
        # Loading user function
        if self.mbs.opt_load_c < 2:
            if __DEBUG__ : print( "DEBUG>>  Loading user functions")
            self.__load_user_fct__(self.user_path)
        # Loading symbolic function
        if self.mbs.opt_load_c < 1:
            if __DEBUG__ : print( "DEBUG>>  Loading symbolic functions")
            self.__load_symbolic_fct__(self.symbolic_path)
        
    def __str__(self):
        if __DEBUG__ : print( "DEBUG>>  start of __str")
        
        return self.module_name + " instance has nothing to be printed from C library!"
    
    def __load_user_fct__(self, user_path):
        """ 
            Load user function where some args depend on MbsInvdyn/MbsSolvekin module.
            
            Load the user functions in which some of the arguments are  
            dependent of MbsInvdyn/MbsSolvekin module instance. The functions will be 
            assigned to the MbsData instance when the 'run' functions is called 
            and unassigned at the end.
            
            The loader user functions are :
             - user_cons_hJ (from user_cons_hJ.py)
             - user_cons_jdqd (from user_cons_jdqd.py)
        """
        
        template_path = os.path.join(__MODULE_DIR__,'../templates/user')
        
        
        # cons_hJ
        user_file = "user_cons_hJ.py"
        path = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        module = imp.load_source(user_file[:-3], path)
        self.user_cons_hJ = module.user_cons_hJ
        
        # cons_jdqd
        user_file = "user_cons_jdqd.py"
        path = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        module = imp.load_source(user_file[:-3], path)
        self.user_cons_jdqd = module.user_cons_jdqd
        return
    
    def __load_symbolic_fct__(self, symbolic_path):
        """ 
            Load symbolic function where some args depend on MbsInvdyn/MbsSolvekin module.
            
            Load the symb functions in which some of the arguments are  
            dependent of MbsInvdyn/MbsSolvekin module instance. The functions will be 
            assigned to the MbsData instance when the 'run' functions is called 
            and unassigned at the end.
            
            The loader user functions are :
             - cons_hJ (from mbs_cons_hJ_MBSNAME.py)
             - cons_jdqd (from mbs_cons_jdqd_MBSNAME.py)
             - invdyna (from mbs_invdyna_MBSNAME.py)
             - dirdyna (from mbs_dirdyna_MBSNAME.py)
        """
        mbs_name = self.mbs.mbs_name
        template_path = os.path.join(__MODULE_DIR__,'../templates/symbolic')
        # mbs_cons_hJ
        symb_file = "mbs_cons_hJ_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_cons_hJ_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_cons_hJ = module.cons_hJ
        
        
        # mbs_cons_jdqd
        symb_file = "mbs_cons_jdqd_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_cons_jdqd_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_cons_jdqd = module.cons_jdqd
        
        # invdyna
        #not used in the MbsSolvekin module
        symb_file = "mbs_invdyna_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_invdyna_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_invdyna = module.invdyna
       
        # dirdyna
        #not used in the MbsInvdyn module
        symb_file = "mbs_dirdyna_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_dirdyna_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_dirdyna = module.dirdyna
        
        return
    
    def __assign_user_fct__(self):
        """ 
            Assign and wrap python user function where some args depend 
            on MbsInvdyn/MbsSolvekin module. Store the functions in the MbsData instance 
            and assign the pointer of functions in the C structure
        """
        # cons_hJ
        self.ptrs_to_user_fcts["user_cons_hJ"] = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.user_cons_hJ(self._huserc, self._Juserc, self.mbs, tsim))
        self.mbs.user_cons_hJ = self.user_cons_hJ
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = self.ptrs_to_user_fcts["user_cons_hJ"]
        
        # cons_jdqd
        self.ptrs_to_user_fcts["user_cons_jdqd"] = user_cons_jdqd_wrap(lambda jdqd,mbs,tsim : self.user_cons_jdqd(self._jdqduserc, self.mbs))
        self.mbs.user_cons_jdqd = self.user_cons_jdqd
        self.mbs.mbs_data_ptr.contents.user_cons_jdqd = self.ptrs_to_user_fcts["user_cons_jdqd"]
        return
    
    def __assign_symbolic_fct__(self):
        """ 
            Assign and wrap python symbolic function where some args depend 
            on MbsInvdyn/MbsSolvekin module. Store the functions in the MbsData instance 
            and assign the pointer of functions in the C structure
        """
        
        # mbs_cons_hJ
        self.ptrs_to_symb_fcts["mbs_cons_hJ"] = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.mbs_cons_hJ(self._h,self._Jac, self.mbs))
        self.mbs.mbs_cons_hJ = self.mbs_cons_hJ
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = self.ptrs_to_symb_fcts["mbs_cons_hJ"]
        
        # mbs_cons_jdqd
        self.ptrs_to_symb_fcts["mbs_cons_jdqd"] = mbs_cons_jdqd_wrap(lambda jdqd,mbs,tsim : self.mbs_cons_jdqd(self._jdqd, self.mbs))
        self.mbs.mbs_cons_jdqd = self.mbs_cons_jdqd
        self.mbs.mbs_data_ptr.contents.mbs_cons_jdqd = self.ptrs_to_symb_fcts["mbs_cons_jdqd"]
        
        # invdyna
        #not used in the MbsSolvekin module
        self.ptrs_to_symb_fcts["mbs_invdyna"] = mbs_invdyna_wrap(lambda phi,mbs,tsim : self.mbs_invdyna(self._phi, self.mbs, tsim))
        self.mbs.mbs_invdyna = self.mbs_invdyna
        self.mbs.mbs_data_ptr.contents.mbs_invdyna = self.ptrs_to_symb_fcts["mbs_invdyna"]
        
        # dirdyna
        #not used in the MbsInvdyn module
        self.ptrs_to_symb_fcts["mbs_dirdyna"] = mbs_dirdyna_wrap(lambda M,c,mbs,tsim : self.mbs_dirdyna(self._M, self._c, self.mbs, tsim))
        self.mbs.mbs_dirdyna = self.mbs_dirdyna
        self.mbs.mbs_data_ptr.contents.mbs_dirdyna = self.ptrs_to_symb_fcts["mbs_dirdyna"]
         
        return
    
    def __unassign_user_fct__(self):
        """ 
            Unassign user function where some args depend on MbsInvdyn/MbsSolvekin module.
        """
        
        self.ptrs_to_user_fcts = {}
        
        self.mbs.user_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("user_cons_hJ"))
        self.mbs.user_cons_jdqd = None
        self.mbs.mbs_data_ptr.contents.user_cons_jdqd = user_cons_jdqd_wrap(lambda h,Jac,mbs,tsim : callback_undefined("user_cons_jdqd"))
        
        return
        
    def __unassign_symbolic_fct__(self):
        
        """ 
            Unassign symbolic function where some args depend on MbsInvdyn/MbsSolvekin module.
        """
        
        self.ptrs_to_symb_fcts = {}
        
        self.mbs.mbs_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("mbs_cons_hJ"))
        self.mbs.mbs_cons_jdqd = None
        self.mbs.mbs_data_ptr.contents.mbs_cons_jdqd = mbs_cons_jdqd_wrap(lambda jdqd,mbs,tsim : callback_undefined("mbs_cons_jdqd"))
        self.mbs.mbs_invdyna = None
        self.mbs.mbs_data_ptr.contents.mbs_invdyna = mbs_invdyna_wrap(lambda Q,mbs,tsim : callback_undefined("mbs_invdyna"))
        self.mbs.mbs_dirdyna = None
        self.mbs.mbs_data_ptr.contents.mbs_dirdyna = mbs_dirdyna_wrap(lambda M,c,mbs,tsim : callback_undefined("mbs_dirdyna"))
        
        return
        
    def __del__(self):
        
        libmodules.mbs_delete_invdyn(self.mbs_invdyn_ptr, self.mbs.mbs_data_ptr)
        if __DEBUG__ : print("DEBUG>>  " +self.module_name + " pointer deleted")
        
    def run(self):
        """
            Run an inverse dynamics analysis
            
            Options can be setted with set_options 
            Options can be retrieved with get_options
            
            Results are stored in the field results (if store_results == True)
        """
        # Assing required user functions
        if self.mbs.opt_load_c < 2:
            self.__assign_user_fct__()
            self.mbs.__assign_user_fct__()
        
        # Assing required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__assign_symbolic_fct__()
            self.mbs.__assign_symb_fct__()
        
        if not self.store_results:   
            libmodules.mbs_run_invdyn(self.mbs_invdyn_ptr, self.mbs.mbs_data_ptr)
        else:
            #save2file forced to 1 because if buffers don't have the complete results, results are loaded from files
            self.set_options(save2file = 1)
            libmodules.mbs_invdyn_init(self.mbs_invdyn_ptr, self.mbs.mbs_data_ptr)
            libmodules.mbs_invdyn_loop(self.mbs_invdyn_ptr, self.mbs.mbs_data_ptr)
       
            # Results (buffer) memory is kept BUT FILES WILL BE WRITTEN LATER
            if self.get_options("save2file"):
               self._load_results()
            
            libmodules.mbs_invdyn_finish(self.mbs_invdyn_ptr, self.mbs.mbs_data_ptr)
        
        # Anassign user functions
        if self.mbs.opt_load_c < 2:
            self.__unassign_user_fct__()
            self.mbs.__unassign_user_fct__()
        
        # Anassing required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__unassign_symbolic_fct__()
            self.mbs.__unassign_symb_fct__()
            
        return self.results
        
    def _load_results(self):
        """
            Load the results from the buffers
            
            If the beginning of the integration is not available in the buffers, 
            the complete results are loaded from files 
        """
        #c_mbs_invdyn_write_buffers(self.mbs_invdyn_ptr)
        size1 = self.mbs_invdyn_ptr.contents.buffers[0].contents.index
        if size1 == 0:
            size1 = self.mbs_invdyn_ptr.contents.buffers[0].contents.size
        size2 = self.mbs_invdyn_ptr.contents.buffers[0].contents.nx+1
        
        # array are initialized to the time pointer so as to start index of joints at 1 (we have to ensure contiguity between t and x in buffers ! ! !)    
        self.results.q = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[0].contents.tx ,(size1, size2)))
        # get time array from the q buffer
        self.results.t = self.results.q[:,0]
        if not self.results.t[0] == self.get_options("t0"):
            print("The beginning of the integration is not available in the buffer.\n The complete results have to be loaded from files.")
            filename = bytes_to_str(ctypes.string_at(self.mbs_invdyn_ptr.contents.buffers[0].contents.filename))
            self.results.load_results_from_file(filename, module = 6)
        # get qd and qdd buffer
        self.results.qd = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[1].contents.tx ,(size1, size2)))
        self.results.qdd = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[2].contents.tx ,(size1, size2)))
        size2 = self.mbs_invdyn_ptr.contents.buffers[3].contents.nx+1
        self.results.Qq = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[3].contents.tx ,(size1, size2)))
        if self.module_name == "MbsInvdyn":
            size2 = self.mbs_invdyn_ptr.contents.buffers[4].contents.nx+1
            self.results.Qa = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[4].contents.tx ,(size1, size2)))
            buffer_id = 5
        else:
            buffer_id = 4
        if self.mbs.Nlink:
            size2 = self.mbs_invdyn_ptr.contents.buffers[buffer_id].contents.nx+1
            self.results.Z = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[buffer_id].contents.tx ,(size1, size2)))
            self.results.Zd = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[buffer_id+1].contents.tx ,(size1, size2)))
            self.results.Fl = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[buffer_id+2].contents.tx ,(size1, size2)))
            buffer_id = buffer_id + 3
        if self.mbs.nqc:
            size2 = self.mbs_invdyn_ptr.contents.buffers[buffer_id].contents.nx+1
            self.results.Qc = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[buffer_id].contents.tx ,(size1, size2)))
            buffer_id = buffer_id + 1
        if self.mbs.nhu:
            size2 = self.mbs_invdyn_ptr.contents.buffers[buffer_id].contents.nx+1
            self.results.Lambda = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.buffers[buffer_id].contents.tx ,(size1, size2)))
            
        if self.mbs_invdyn_ptr.contents.user_buffer.contents.nx:
            size = self.mbs_invdyn_ptr.contents.user_buffer.contents.index
            nbOutput = self.mbs_invdyn_ptr.contents.user_buffer.contents.nx
            user_out = np.copy(np.ctypeslib.as_array(self.mbs_invdyn_ptr.contents.user_buffer.contents.X[0],(nbOutput, size)))
            for i in range(nbOutput):
                name = bytes_to_str(self.mbs_invdyn_ptr.contents.user_buffer.contents.names[i])
                self.results.outputs[name] = user_out[i,:]
        
    def set_options(self, **kwargs):
        """
        Set the specified options for Invdyn module
        
        Parameters
        ----------
        
        motion : int 
            Determines how the evolution of q, qd qdd are provided.
            oneshot (default) : Uses q, qd and qdd at the current time and configuration of the MBS.
            closeloop : Only for inverse kinematics analysis. Same as oneshot but save the iteration of the close loop process (NR).
            trajectory : Compute the evolution of between t0 and tf.
        trajectoryqname : char 
            Give the path and filename (with extention) to the file containing the input trajectory in position of the joints.
            Only required for motion == 'trajectory' if some joints are independant.
            By default, pointer to NULL.
            Requirement on the file:
            First column is the time, other are the joints coordinates.
            Either provide only the independant joint (by increasing index), user_drivenjoint will be called.
            Either provide all the joint, user_drivenjoint will be neglected.
        trajectoryqname : char 
            Give the path and filename (with extention) to the file containing the input trajectory in velocity of the joints.
            Only required for motion == 'trajectory' if some joints are independant.
            By default, pointer to NULL.
            Requirement on the file:
            First column is the time, other are the joints velocities.
            Either provide only the independant joint (by increasing index), user_drivenjoint will be called.
            Either provide all the joint, user_drivenjoint will be neglected.
        trajectoryqname : char 
            Give the path and filename (with extention) to the file containing the input trajectory in acceleration of the joints.
            Only required for motion == 'trajectory' if some joints are independant.
            By default, pointer to NULL.
            Requirement on the file:
            First column is the time, other are the joints accelerations.
            Either provide only the independant joint (by increasing index), user_drivenjoint will be called.
            Either provide all the joint, user_drivenjoint will be neglected.
        t0 : float
            initial time of the simulation, default = 0.0
        tf : float
            final time of the simulation, default = 5.0
        dt0 : float
            Time step, default = 0.001
            For motion == 'trajectory' (for inverse kinematic analysis, there must be at least one independent joint):
            Set to 0.0 to use the time vector of the coordinate input file (MbsSolvekinOptions::t0 and MbsSolvekinOptions::tf are ignored).
            Specify a value to use specific time vecor.
        save2file : int
            Determine whether results are written to files on disk (in MbsSolvekinOptions::respath folder):  
            1: results are saved
            0: results are not saved 
            default = 1
        resfilename : str
            The keyword used for determining the name of result files
        respath : str
            Path in which result file are saved.
            Default: the resultsR folder of the project
        animpath : str
            Path in which anim file is saved.
            Default: the animationR folder of the project
        save_anim : int
            1 to save the anim file
            0 otherwise (only valid if 'save2file' is set to 1)
            default = 1
        framerate : int
            number of frame per second for the .anim file
            default = 1000
        saveperiod : int
            The number of time steps between two buffer records
            default: 1 (every time step are recorded)
        max_save_user : int
            The maximal number of user variables saved
            default: 12
        buffersize : int 
            The number of time step that can be recorded in the buffer.
            Results are written to disk when the buffer is full.
            default: -1
            compute the buffer size for saving results only once at the end according to dt0, t0 and tf.
        verbose : int
            Gives informations during the process.
            default: 1, to disable set it to 0.
        store_results : boolean
            1 to save a copy of the results from the buffers
            default = 1
            
        """
        
        for key, value in kwargs.items():
            if key == "motion":
                if (type(value) is int) and (value>=0 and value<=2):
                    self.mbs_invdyn_ptr.contents.options.contents.motion = value
                elif type(value) is str:
                    if value == "oneshot":
                        self.mbs_invdyn_ptr.contents.options.contents.motion = 0
                    elif value == "closeloop":
                        self.mbs_invdyn_ptr.contents.options.contents.motion = 1
                    elif value == "trajectory":
                        self.mbs_invdyn_ptr.contents.options.contents.motion = 2
                    else:
                        print(">>" +self.module_name + ">>  " + value + " is not an motion")
                else:
                    print(">>" +self.module_name + ">>  " + value + " is not an motion")
            elif key == "trajectoryqname":
                self.mbs_invdyn_ptr.contents.options.contents.trajectoryqname = str_to_bytes(value)
            elif key == "trajectoryqdname":
                self.mbs_invdyn_ptr.contents.options.contents.trajectoryqdname = str_to_bytes(value)
            elif key == "trajectoryqddname":
                self.mbs_invdyn_ptr.contents.options.contents.trajectoryqddname = str_to_bytes(value)
            elif key == "t0":
                self.mbs_invdyn_ptr.contents.options.contents.t0 = value
            elif key == "tf":
                self.mbs_invdyn_ptr.contents.options.contents.tf = value
            elif key == "dt":
                self.mbs_invdyn_ptr.contents.options.contents.dt = value
            elif key == "save2file":
                self.mbs_invdyn_ptr.contents.options.contents.save2file = value
            elif key == "resfilename":
                self.mbs_invdyn_ptr.contents.options.contents.resfilename = str_to_bytes(value)
            elif key == "respath":
                self.mbs_invdyn_ptr.contents.options.contents.respath = str_to_bytes(value)
            elif key == "animpath":
                self.mbs_invdyn_ptr.contents.options.contents.animpath = str_to_bytes(value)
            elif key == "save_anim":
                self.mbs_invdyn_ptr.contents.options.contents.save_anim = value
            elif key == "framerate":
                self.mbs_invdyn_ptr.contents.options.contents.framerate = value
            elif key == "saveperiod":
                self.mbs_invdyn_ptr.contents.options.contents.saveperiod= value
            elif key == "max_save_user":
                self.mbs_invdyn_ptr.contents.options.contents.max_save_user = value
            elif key == "buffersize":
                self.mbs_invdyn_ptr.contents.options.contents.buffersize = value
            elif key == "verbose":
                self.mbs_invdyn_ptr.contents.options.contents.verbose = value
            elif key == "store_results":
                self.store_results = value
            else:
                print(">>" +self.module_name + ">> The option "+key+ " is not defined in this module")
                
    def get_options(self, *args):
        """
        Get the specified options for Invdyn module
        
        Parameters
        ----------
        
        The different options specifed in the documentation of set_options
        
        Returns
        -------
    
        The value of the different options specifed in the documentation of set_options
        
        
        """
        options = []
        for key in args:
            if key == "motion":
                if self.mbs_invdyn_ptr.contents.options.contents.motion == 0:
                    options.append("oneshot")
                elif self.mbs_invdyn_ptr.contents.options.contents.motion == 1:
                    options.append("closeloop")
                elif self.mbs_invdyn_ptr.contents.options.contents.motion == 2:
                    options.append("trajectory")
            elif key == "trajectoryqname":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_invdyn_ptr.contents.options.contents.trajectoryqname)))
            elif key == "trajectoryqdname":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_invdyn_ptr.contents.options.contents.trajectoryqdname)))
            elif key == "trajectoryqddname":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_invdyn_ptr.contents.options.contents.trajectoryqddname)))
            elif key == "t0":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.t0)
            elif key == "tf":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.tf)
            elif key == "dt":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.dt)
            elif key == "save2file":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.save2file)
            elif key == "resfilename":
                if self.mbs_invdyn_ptr.contents.options.contents.resfilename is None:
                    print(">>" +self.module_name + ">> The option "+key+ " is not assigned")
                else:
                    options.append(bytes_to_str(ctypes.string_at(self.mbs_invdyn_ptr.contents.options.contents.resfilename)))
                
            elif key == "respath":
                if self.mbs_invdyn_ptr.contents.options.contents.respath is None:
                    print(">>" +self.module_name + ">> The option "+key+ " is not assigned")
                else:
                    options.append(bytes_to_str(ctypes.string_at(self.mbs_invdyn_ptr.contents.options.contents.respath)))
            elif key == "animpath":
                if self.mbs_invdyn_ptr.contents.options.contents.animpath is None:
                    print(">>" +self.module_name + ">> The option "+key+ " is not assigned")
                else:
                    options.append(bytes_to_str(ctypes.string_at(self.mbs_invdyn_ptr.contents.options.contents.animpath)))
            elif key == "save_anim":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.save_anim)
            elif key == "framerate":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.framerate)
            elif key == "saveperiod":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.saveperiod)
            elif key == "max_save_user":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.max_save_user)
            elif key == "buffersize":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.buffersize)
            elif key == "verbose":
                options.append(self.mbs_invdyn_ptr.contents.options.contents.verbose)
            elif key == "store_results":
                options.append(self.store_results)
            else:
                print(">>" +self.module_name + ">> The option "+key+ " is not defined in this module")
                
        if len(options) == 0:
            return
        if len(options) == 1:
            return options[0]
            
        return tuple(options)