#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
-------------------------------
(c) Universite catholique de Louvain, 2019

Creation : 2019 by LB
Last update : 2019

version MBsysC v1.11.2

-------------------------------
Portable Python interface to MBsysC using Ctypes.

Define the class MbsModal based on the MbsModal structure of MBsysC.
"""

import os
import imp
import ctypes
import numpy as np

# importing MbsysPy functions
from ..mbs_utilities import bytes_to_str
from ..mbs_utilities import str_to_bytes

# importing libraries
from .._mbsysc_loader.loadlibs import libmodules

# importing utilities function
from ..mbs_utilities import callback_undefined

# importing wrapping function
from .._mbsysc_loader.callback import user_cons_hJ_wrap
from .._mbsysc_loader.callback import user_cons_jdqd_wrap

from .._mbsysc_loader.callback import mbs_cons_hJ_wrap
from .._mbsysc_loader.callback import mbs_cons_jdqd_wrap
from .._mbsysc_loader.callback import mbs_invdyna_wrap
from .._mbsysc_loader.callback import mbs_dirdyna_wrap

#==============================================================================
# Global parameter of the current module
#==============================================================================
__DEBUG__ = False
__MODULE_DIR__ = os.path.dirname(os.path.abspath(__file__))


#==============================================================================
# Defining Python MbsModal class
#==============================================================================

class MbsModal(object):
    
    def __init__(self, mbs, user_path=None, symbolic_path=None):
        self.module_name = type(self).__name__
        
        if __DEBUG__ : print( "DEBUG>>  Creating MbsModal struct for "+mbs.mbs_name+"' MBS.")
        self.mbs_modal_ptr = libmodules.mbs_new_modal(mbs.mbs_data_ptr)
        if __DEBUG__ : print( "DEBUG>>  MbsModal structure loaded")
        self.mbs = mbs
        
        
        self.user_path = self.mbs.user_path
        if not (user_path==None):
            project_path  = bytes_to_str(self.mbs.project_path)
            user_path = os.path.join(project_path, user_path)
            # Error handeling
            if not os.path.isdir(user_path):
                print ('The user function directory for direct dynamic module does not exist: "'+ user_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.user_path+'".')
            else:
                self.user_path = user_path
        # Path to user function used by partitionning modue
        self.symbolic_path = self.mbs.symbolic_path
        if not (user_path==None):
            project_path  = bytes_to_str(self.mbs.project_path)
            symbolic_path = os.path.join(project_path, symbolic_path)
            # Error handeling
            if not os.path.isdir(symbolic_path):
                print ('Thesymbolic function directory for direct dynamic module does not exist: "'+ symbolic_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.symbolic_path+'".')
            else:
                self.symbolic_path = symbolic_path
        
        # dictionary containing the pointers to avoid garbage collecting
        self.ptrs_to_user_fcts = dict()
        self.ptrs_to_symb_fcts = dict()
        
        # Storing project function pointer
        self.user_cons_hJ = None
        self.user_cons_jdqd = None
        self.mbs_cons_hJ = None
        self.mbs_cons_jdqd = None
        self.mbs_invdyna = None
        
        # Storing Results
        self.results = MbsResultModal()
        self.store_results = False
        
        # Exposing some memory
        if __DEBUG__ : print( "DEBUG>>  Exposing " +self.module_name + " fields")
        # Constraints
        self._h = self._Jac = self._jdqd = None
        if self.mbs.Ncons>0:
            self._h = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.h, (self.mbs.Ncons+1,))
            self._Jac = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.Jac[0], (self.mbs.Ncons+1,self.mbs.njoint+1))
            self._jdqd = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.jdqd, (self.mbs.Ncons+1,))
        self._huserc = self._Juserc = self._jdqduserc = None
        if self.mbs.Nuserc>0:
            self._huserc = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.huserc, (self.mbs.Nuserc+1,))
            self._Juserc = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.Juserc[0], (self.mbs.Nuserc+1,self.mbs.njoint+1))
            self._jdqduserc = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.jdqduserc, (self.mbs.Nuserc+1,))
        # Fields from equation of motion
        self._M = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.M[0], (self.mbs.njoint+1,self.mbs.njoint+1))
        self._c = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.c, (self.mbs.njoint+1,))
        self._F = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.F, (self.mbs.njoint+1,))
        self._phi = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.aux.contents.phi, (self.mbs.njoint+1,))
        
        # Loading user function
        if self.mbs.opt_load_c < 2:
            if __DEBUG__ : print( "DEBUG>>  Loading user functions")
            self.__load_user_fct__(self.user_path)
        # Loading symbolic function
        if self.mbs.opt_load_c < 1:
            if __DEBUG__ : print( "DEBUG>>  Loading symbolic functions")
            self.__load_symbolic_fct__(self.symbolic_path)
        
    def __str__(self):
        if __DEBUG__ : print( "DEBUG>>  start of __str")
        
        return self.module_name + " instance has nothing to be printed from C library!"
    
    def __load_user_fct__(self, user_path):
        """ 
            Load user function where some args depend on MbsModal module.
            
            Load the user functions in which some of the arguments are  
            dependent of MbsModal module instance. The functions will be 
            assigned to the MbsData instance when the 'run' functions is called 
            and unassigned at the end.
            
            The loader user functions are :
             - user_cons_hJ (from user_cons_hJ.py)
             - user_cons_jdqd (from user_cons_jdqd.py)
        """
        
        template_path = os.path.join(__MODULE_DIR__,'../templates/user')
        
        
        # cons_hJ
        user_file = "user_cons_hJ.py"
        path = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        module = imp.load_source(user_file[:-3], path)
        self.user_cons_hJ = module.user_cons_hJ
        
        # cons_jdqd
        user_file = "user_cons_jdqd.py"
        path = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        module = imp.load_source(user_file[:-3], path)
        self.user_cons_jdqd = module.user_cons_jdqd
        return
    
    def __load_symbolic_fct__(self, symbolic_path):
        """ 
            Load symbolic function where some args depend on MbsModal module.
            
            Load the symb functions in which some of the arguments are  
            dependent of MbsModal module instance. The functions will be 
            assigned to the MbsData instance when the 'run' functions is called 
            and unassigned at the end.
            
            The loader user functions are :
             - cons_hJ (from mbs_cons_hJ_MBSNAME.py)
             - cons_jdqd (from mbs_cons_jdqd_MBSNAME.py)
             - invdyna (from mbs_invdyna_MBSNAME.py)
             - dirdyna (from mbs_dirdyna_MBSNAME.py)
        """
        mbs_name = self.mbs.mbs_name
        template_path = os.path.join(__MODULE_DIR__,'../templates/symbolic')
        # mbs_cons_hJ
        symb_file = "mbs_cons_hJ_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_cons_hJ_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_cons_hJ = module.cons_hJ
        
        
        # mbs_cons_jdqd
        symb_file = "mbs_cons_jdqd_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_cons_jdqd_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_cons_jdqd = module.cons_jdqd
        
        # invdyna
        symb_file = "mbs_invdyna_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_invdyna_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_invdyna = module.invdyna
       
        # dirdyna
        symb_file = "mbs_dirdyna_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_dirdyna_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_dirdyna = module.dirdyna
        
        return
    
    def __assign_user_fct__(self):
        """ 
            Assign and wrap python user function where some args depend 
            on MbsModal module. Store the functions in the MbsData instance 
            and assign the pointer of functions in the C structure
        """
        # cons_hJ
        self.ptrs_to_user_fcts["user_cons_hJ"] = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.user_cons_hJ(self._huserc, self._Juserc, self.mbs, tsim))
        self.mbs.user_cons_hJ = self.user_cons_hJ
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = self.ptrs_to_user_fcts["user_cons_hJ"]
        
        # cons_jdqd
        self.ptrs_to_user_fcts["user_cons_jdqd"] = user_cons_jdqd_wrap(lambda jdqd,mbs,tsim : self.user_cons_jdqd(self._jdqduserc, self.mbs))
        self.mbs.user_cons_jdqd = self.user_cons_jdqd
        self.mbs.mbs_data_ptr.contents.user_cons_jdqd = self.ptrs_to_user_fcts["user_cons_jdqd"]
        return
    
    def __assign_symbolic_fct__(self):
        """ 
            Assign and wrap python symbolic function where some args depend 
            on MbsModal module. Store the functions in the MbsData instance 
            and assign the pointer of functions in the C structure
        """
        
        # mbs_cons_hJ
        self.ptrs_to_symb_fcts["mbs_cons_hJ"] = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.mbs_cons_hJ(self._h,self._Jac, self.mbs))
        self.mbs.mbs_cons_hJ = self.mbs_cons_hJ
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = self.ptrs_to_symb_fcts["mbs_cons_hJ"]
        
        # mbs_cons_jdqd
        self.ptrs_to_symb_fcts["mbs_cons_jdqd"] = mbs_cons_jdqd_wrap(lambda jdqd,mbs,tsim : self.mbs_cons_jdqd(self._jdqd, self.mbs))
        self.mbs.mbs_cons_jdqd = self.mbs_cons_jdqd
        self.mbs.mbs_data_ptr.contents.mbs_cons_jdqd = self.ptrs_to_symb_fcts["mbs_cons_jdqd"]
        
        # invdyna
        self.ptrs_to_symb_fcts["mbs_invdyna"] = mbs_invdyna_wrap(lambda phi,mbs,tsim : self.mbs_invdyna(self._phi, self.mbs, tsim))
        self.mbs.mbs_invdyna = self.mbs_invdyna
        self.mbs.mbs_data_ptr.contents.mbs_invdyna = self.ptrs_to_symb_fcts["mbs_invdyna"]
        
        # dirdyna
        self.ptrs_to_symb_fcts["mbs_dirdyna"] = mbs_dirdyna_wrap(lambda M,c,mbs,tsim : self.mbs_dirdyna(self._M, self._c, self.mbs, tsim))
        self.mbs.mbs_dirdyna = self.mbs_dirdyna
        self.mbs.mbs_data_ptr.contents.mbs_dirdyna = self.ptrs_to_symb_fcts["mbs_dirdyna"]
         
        return
    
    def __unassign_user_fct__(self):
        """ 
            Unassign user function where some args depend on MbsModal module.
        """
        
        self.ptrs_to_user_fcts = {}
        
        self.mbs.user_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("user_cons_hJ"))
        self.mbs.user_cons_jdqd = None
        self.mbs.mbs_data_ptr.contents.user_cons_jdqd = user_cons_jdqd_wrap(lambda h,Jac,mbs,tsim : callback_undefined("user_cons_jdqd"))
        
        return
        
    def __unassign_symbolic_fct__(self):
        
        """ 
            Unassign symbolic function where some args depend on MbsModal module.
        """
        
        self.ptrs_to_symb_fcts = {}
        
        self.mbs.mbs_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("mbs_cons_hJ"))
        self.mbs.mbs_cons_jdqd = None
        self.mbs.mbs_data_ptr.contents.mbs_cons_jdqd = mbs_cons_jdqd_wrap(lambda jdqd,mbs,tsim : callback_undefined("mbs_cons_jdqd"))
        self.mbs.mbs_invdyna = None
        self.mbs.mbs_data_ptr.contents.mbs_invdyna = mbs_invdyna_wrap(lambda Q,mbs,tsim : callback_undefined("mbs_invdyna"))
        self.mbs.mbs_dirdyna = None
        self.mbs.mbs_data_ptr.contents.mbs_dirdyna = mbs_dirdyna_wrap(lambda M,c,mbs,tsim : callback_undefined("mbs_dirdyna"))
        
        return
        
    def __del__(self):
        
        libmodules.mbs_delete_modal(self.mbs_modal_ptr, self.mbs.mbs_data_ptr)
        if __DEBUG__ : print("DEBUG>>  " +self.module_name + " pointer deleted")
        
    def run(self):
        """
            Run a modal analysis
            
            Options can be setted with set_options 
            Options can be retrieved with get_options
            
        """
        # Assing required user functions
        if self.mbs.opt_load_c < 2:
            self.__assign_user_fct__()
            self.mbs.__assign_user_fct__()
        
        # Assing required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__assign_symbolic_fct__()
            self.mbs.__assign_symb_fct__()
          
        libmodules.mbs_run_modal(self.mbs_modal_ptr, self.mbs.mbs_data_ptr)
        if self.store_results:
            self.load_results()
        
        # Anassign user functions
        if self.mbs.opt_load_c < 2:
            self.__unassign_user_fct__()
            self.mbs.__unassign_user_fct__()
        
        # Anassing required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__unassign_symbolic_fct__()
            self.mbs.__unassign_symb_fct__()
        
        return self.results
        
        
    def load_results(self):
        
        self.results.eval_a = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.eval_a, (self.nx,))
        self.results.eval_b = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.eval_b, (self.nx,))
        
        self.results.evec_a = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.evec_a[0], (self.nx,self.nx))
        self.results.evec_b = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.evec_b[0], (self.nx,self.nx))
        
        self.results.evec_r = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.evec_r[0], (self.nx,self.nx))
        self.results.evec_phi = np.ctypeslib.as_array(self.mbs_modal_ptr.contents.evec_phi[0], (self.nx,self.nx))
        
        
    def set_options(self, **kwargs):
        """
        Set the specified options for Modal module
        
        Parameters
        ----------
        
        save2file : int
            Determine whether results are written to files on disk (in MbsModalOptions::respath folder):  
            1: results are saved
            0: results are not saved 
            default = 1
        resfilename : str
            The keyword used for determining the name of result files
        respath : str
            Path in which result file are saved.
            Default: the resultsR folder of the project
        animpath : str
            Path in which anim file is saved.
            Default: the animationR folder of the project
        save_anim : int
            1 to save the anim file
            0 otherwise (only valid if 'save2file' is set to 1)
            default = 1
        save_mat : int 
            no = 0, yes = 1
            flag to save the linearized matrix A,Mr,Gr,Kr
            default = 0
        save_eval : int
            no = 0, yes = 1
            flag to save the eigen values in two separate files: "_a" for Real part and "_b" for Imag part
            default = 0
        save_evec : int 
            no = 0, yes = 1
            flag to save the eigen vector in two separate files: 
            "_a" for Real part, "_b" for Imag part, "_r" for magnitude and "_phi" for phase
            default = 0
        save_result : int
            no = 0, yes = 1
            flag to save the modal results summary
            default = 1
        save_anim : int 
            no = 0, yes = 1
            flag to save and generate the animation for the modes
            default = 0
        mode_ampl : float 
            amplification factor for the animation of the modes
            application dependent
            default = 0.2
        compute_uxd : int 
            no = 0, yes = 1
            flag to compute the modes associated with the extra constitutive differential equations
            default = 1
        compute_JS : int 
            no = 0, yes = 1
            flag to compute the non diagonal terms
            default = 0 
            TO IMPLEMENT !
        norm_evec : int
            no = 0, yes = 1
            flag to normalize the eigenvectors
            default = 1
        lpk_itermax : int 
            options for the lpk gradient
            default = 10
        lpk_relincr : float 
            options for the lpk gradient
            default = 1e-2
        lpk_absincr : float 
            options for the lpk gradient
            default = 1e-3
        lpk_equitol : float 
            options for the lpk gradient
            default = 1e-6
        lpk_lintol : float 
            options for the lpk gradient
            default = 1e-3
        anim_speed : int 
            flag to decide whether the independent speeds have to be integrated while generating the animation files
            default = 0
        anim_dt : int
            delta t for the animation generation
            default = 1e-2
        anim_t : int 
            time for the animation generation
            default = 2
        verbose : int
            Gives informations during the process.
            default: 1, to disable set it to 0.
            
        """
        
        for key, value in kwargs.items():
            if key == "save2file":
                self.mbs_modal_ptr.contents.options.contents.save2file = value
            elif key == "resfilename":
                self.mbs_modal_ptr.contents.options.contents.resfilename = str_to_bytes(value)
            elif key == "respath":
                self.mbs_modal_ptr.contents.options.contents.respath = str_to_bytes(value)
            elif key == "animpath":
                self.mbs_modal_ptr.contents.options.contents.animpath = str_to_bytes(value)
            elif key == "save_anim":
                self.mbs_modal_ptr.contents.options.contents.save_anim = value
            elif key == "save_mat":
                self.mbs_modal_ptr.contents.options.contents.save_mat = value
            elif key == "save_eval":
                self.mbs_modal_ptr.contents.options.contents.save_eval = value
            elif key == "save_evec":
                self.mbs_modal_ptr.contents.options.contents.save_evec = value
            elif key == "save_result":
                self.mbs_modal_ptr.contents.options.contents.save_result = value
            elif key == "mode_ampl":
                self.mbs_modal_ptr.contents.options.contents.mode_ampl = value
            elif key == "compute_uxd":
                self.mbs_modal_ptr.contents.options.contents.compute_uxd = value
            elif key == "compute_JS":
                self.mbs_modal_ptr.contents.options.contents.compute_JS = value
            elif key == "norm_evec":
                self.mbs_modal_ptr.contents.options.contents.norm_evec = value
            elif key == "lpk_itermax":
                self.mbs_modal_ptr.contents.options.contents.lpk_itermax = value
            elif key == "lpk_relincr":
                self.mbs_modal_ptr.contents.options.contents.lpk_relincr = value
            elif key == "lpk_absincr":
                self.mbs_modal_ptr.contents.options.contents.lpk_absincr = value
            elif key == "lpk_equitol":
                self.mbs_modal_ptr.contents.options.contents.lpk_equitol = value
            elif key == "lpk_lintol":
                self.mbs_modal_ptr.contents.options.contents.lpk_lintol = value
            elif key == "anim_speed":
                self.mbs_modal_ptr.contents.options.contents.anim_speed = value
            elif key == "anim_dt":
                self.mbs_modal_ptr.contents.options.contents.anim_dt = value
            elif key == "anim_t":
                self.mbs_modal_ptr.contents.options.contents.anim_t = value
            elif key == "verbose":
                self.mbs_modal_ptr.contents.options.contents.verbose = value
            else:
                print(">>" +self.module_name + ">> The option "+key+ " is not defined in this module")
                
                
    def get_options(self, *args):
        """
        Get the specified options for Modal module
        
        Parameters
        ----------
        
        The different options specifed in the documentation of set_options
        
        Returns
        -------
    
        The value of the different options specifed in the documentation of set_options
        
        
        """
        options = []
        for key in args:
            if key == "save2file":
                options.append(self.mbs_modal_ptr.contents.options.contents.save2file)
            elif key == "resfilename":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_modal_ptr.contents.options.contents.resfilename)))
            elif key == "respath":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_modal_ptr.contents.options.contents.respath)))
            elif key == "animpath":
                options.append(bytes_to_str(ctypes.string_at(self.mbs_modal_ptr.contents.options.contents.animpath)))
            elif key == "save_anim":
                options.append(self.mbs_modal_ptr.contents.options.contents.save_anim)
            elif key == "save_mat":
                options.append(self.mbs_modal_ptr.contents.options.contents.save_mat)
            elif key == "save_eval":
                options.append(self.mbs_modal_ptr.contents.options.contents.save_eval)
            elif key == "save_evec":
                options.append(self.mbs_modal_ptr.contents.options.contents.save_evec)
            elif key == "save_result":
                options.append(self.mbs_modal_ptr.contents.options.contents.save_result)
            elif key == "mode_ampl":
                options.append(self.mbs_modal_ptr.contents.options.contents.mode_ampl)
            elif key == "compute_uxd":
                options.append(self.mbs_modal_ptr.contents.options.contents.compute_uxd)
            elif key == "compute_JS":
                options.append(self.mbs_modal_ptr.contents.options.contents.compute_JS)
            elif key == "norm_evec":
                options.append(self.mbs_modal_ptr.contents.options.contents.norm_evec)
            elif key == "lpk_itermax":
                options.append(self.mbs_modal_ptr.contents.options.contents.lpk_itermax)
            elif key == "lpk_relincr":
                options.append(self.mbs_modal_ptr.contents.options.contents.lpk_relincr)
            elif key == "lpk_absincr":
                options.append(self.mbs_modal_ptr.contents.options.contents.lpk_absincr)
            elif key == "lpk_equitol":
                options.append(self.mbs_modal_ptr.contents.options.contents.lpk_equitol)
            elif key == "lpk_lintol":
                options.append(self.mbs_modal_ptr.contents.options.contents.lpk_lintol)
            elif key == "anim_speed":
                options.append(self.mbs_modal_ptr.contents.options.contents.anim_speed)
            elif key == "anim_dt":
                options.append(self.mbs_modal_ptr.contents.options.contents.anim_dt)
            elif key == "anim_t":
                options.append(self.mbs_modal_ptr.contents.options.contents.anim_t)
            elif key == "verbose":
                options.append(self.mbs_modal_ptr.contents.options.contents.verbose)
            else:
                print(">>" +self.module_name + ">> The option "+key+ " is not defined in this module")
                
        if len(options) == 0:
            return
        if len(options) == 1:
            return options[0]
            
        return tuple(options)
    
            
    #==========================================================================
    # Defining properties
    #==========================================================================
    @property 
    def nx(self):
        return self.mbs_modal_ptr.contents.nx
  
    
    
class MbsResultModal(object):
    
    def __init__(self):
        
        self.eval_a = None
        self.eval_b = None 
        self.evec_a = None
        self.evec_b = None 
        self.evec_r = None 
        self.evec_phi = None
        
        
        