# -*- coding: utf-8 -*-
"""
-------------------------------
(c) Universite catholique de Louvain, 2019

Creation : 2019 by O. Lantsoght
Last update : 2019

version MBsysC v1.11.2

-------------------------------
Portable Python interface to MBsysC using Ctypes.

Define the class MbsPart based on the MbsPart structure of MBsysC.
This class has the functions required to manipulate the coordinate 
partitionning module (options, run, delete)
"""

import os
import imp

import numpy as np

# importing MbsysPy classes

# importing libraries
from .._mbsysc_loader.loadlibs import libmodules

# importing utilities function
from ..mbs_utilities import callback_undefined

# importing wrapping function
from .._mbsysc_loader.callback import user_cons_hJ_wrap

from .._mbsysc_loader.callback import mbs_cons_hJ_wrap


#==============================================================================
# Global parameter of the current module
#==============================================================================
__DEBUG__ = False
__MODULE_DIR__ = os.path.dirname(os.path.abspath(__file__))


#==============================================================================
# Defining Python MbsPart class
#==============================================================================

class MbsPart(object):
    """
    Class of the coordinate partitioning module.
    
    
    Attributes
    ----------
    n_qu : int
        Number of independent variable needed.
    n_qv: int
        Number of dependent variable needed.
    n_hu : int
        Number of independent constraint.
    n_hv : int 
        Number of redundant constraint
    q_closed : ndarray
        Array with the generalized coordinate in closed configuration 
        (constraints solved)
    
    Examples
    --------
    >>> mbs_data = MBsysPy.MbsData("../dataR/ExampleProject.mbs")
    >>> mbs_data.set_qv(2)
    >>> mbs_part = MBsysPy.MbsPart(mbs_data)
    >>> mbs_part.set_options(rowperm=1)
    >>> mbs_part.run()
    
    """
    def __init__(self, mbs, user_path=None, symbolic_path=None):
        """
        Create an instance of the MbsPart class for the provided MbsData instance.

        Parameters
        ----------
        mbs : MbsData
            Instance of MbsData related to the analysis.
        user_path : str or None, optionnal
            The path to the folder containing the user functions. If not provided 
            ('None') the path is retrieved from the MbsData instance 'mbs'.
            default is None
        symbolic_path : str or None, optionnal
            The path to the folder containing the symbolic functions. If not 
            provided ('None') the path is retrieved from the MbsData instance 'mbs'.
            default is None
            
        Returns
        -------
            A MbsPart instance.
        """
        if __DEBUG__ : print( "DEBUG>>  Creating MbsPart struct for "+mbs.mbs_name+"' MBS.")
        self.mbs_part_ptr = libmodules.mbs_new_part(mbs.mbs_data_ptr)
        if __DEBUG__ : print( "DEBUG>>  MbsPartmbs_ structure loaded")
        self.mbs = mbs
        
        # Path to user function used by partitionning modue
        self.user_path = self.mbs.user_path
        if not (user_path==None):
            project_path  = self.mbs.project_path
            user_path = os.path.join(project_path, user_path)
            # Error handeling
            if not os.path.isdir(user_path):
                print ('The partitionning user function directory does not exist: "'+ user_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.user_path+'".')
            else:
                self.user_path = user_path
        # Path to user function used by partitionning modue
        self.symbolic_path = self.mbs.symbolic_path
        if not (user_path==None):
            project_path  = self.mbs.project_path
            symbolic_path = os.path.join(project_path, symbolic_path)
            # Error handeling
            if not os.path.isdir(symbolic_path):
                print ('The partitionning symbolic function directory does not exist: "'+ symbolic_path+'"')
                print ('The current root folder is: "'+os.getcwd()+'"')
                print ('The following directory is used instead: "'+self.symbolic_path+'".')
            else:
                self.symbolic_path = symbolic_path
        
        # pointers list to avoid garbage collecting
        self.ptrs_to_user_fcts = dict()
        self.ptrs_to_symb_fcts = dict()
        
        # Storing project function pointer
        self.user_cons_hJ = None
        self.mbs_cons_hJ = None
        
        # Exposing some memory
        if __DEBUG__ : print( "DEBUG>>  Exposing MbsPart fields")
        self.__q_closed = np.ctypeslib.as_array(self.mbs_part_ptr.contents.q_closed, (self.mbs.njoint,))
        
        # Loading user function
        if self.mbs.opt_load_c < 2:
            if __DEBUG__ : print( "DEBUG>>  Loading user functions")
            self.__load_user_fct__(self.user_path)
            if self.mbs.opt_load_c < 1:
                # Loading symbolic function
                if __DEBUG__ : print( "DEBUG>>  Loading symbolic functions")
                self.__load_symbolic_fct__(self.symbolic_path)

    def __str__(self):
        if __DEBUG__ : print( "DEBUG>>  start of __str")
        
        return "MbsPart instance has nothing to be printed from C library!"
        
    def __load_user_fct__(self, user_path):
        """ 
        Load user function where some args depend on MbsPart module.
        
        Load the user functions in which some of the arguments are  
        dependent of MbsPart module instance. The functions will be 
        assigned to the MbsData instance when the 'run' function is called 
        and unassigned at the end.
        
        The loader user functions are :
         - user_cons_hJ (from user_cons_hJ.py)

        Parameters
        ----------
        user_path : str
            The path to the folder containing the user functions.
        """
        template_path = os.path.join(__MODULE_DIR__,'../templates/user')
        
        # cons_hJ
        user_file = "user_cons_hJ.py"
        path = os.path.abspath(os.path.join(user_path, user_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+user_file+"' not found in folder '"+os.path.dirname(path))
            path = os.path.abspath(os.path.join(template_path, user_file))
        module = imp.load_source(user_file[:-3], path)
        self.user_cons_hJ = module.user_cons_hJ
        return
        
    def __load_symbolic_fct__(self, symbolic_path):
        """ 
        Load symbolic function where some args depend on MbsPart module.
        
        Load the symb functions in which some of the arguments are  
        dependent of MbsPart module instance. The functions will be 
        assigned to the MbsData instance when the 'run' function is called 
        and unassigned at the end.
        
        The loader user functions are :
         - cons_hJ (from mbs_cons_hJ_MBSNAME.py)

        Parameters
        ----------
        symbolic_path : str
            The path to the folder containing the symbolic functions.
        """
        mbs_name = self.mbs.mbs_name
        template_path = os.path.join(__MODULE_DIR__,'../templates/symbolic')
        # mbs_cons_hJ
        symb_file = "mbs_cons_hJ_"+mbs_name+".py"
        path = os.path.abspath(os.path.join(symbolic_path, symb_file))
        if not os.path.isfile(path):
            if __DEBUG__ : print( "DEBUG>>  file '"+symb_file+"' not found in folder '"+os.path.dirname(path))
            symb_file = "mbs_cons_hJ_PRJ.py"
            path = os.path.abspath(os.path.join(template_path, symb_file))
        module = imp.load_source(symb_file[:-3], path)
        self.mbs_cons_hJ = module.cons_hJ
        return
    
    def __assign_user_fct__(self):
        """ 
        Assign and wrap python user function where some args depend on MbsPart 
        module. Store the functions in the MbsData instance and assign the 
        pointer of functions in the C structure.
            
        """
        
        self.ptrs_to_user_fcts["user_cons_hJ"] = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.__callback_user_cons_hJ(self.user_cons_hJ,h,Jac,tsim))
        self.mbs.user_cons_hJ = self.user_cons_hJ
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = self.ptrs_to_user_fcts["user_cons_hJ"]
        return
        
    def __assign_symbolic_fct__(self):
        """ 
        Assign and wrap python symbolic function where some args depend on MbsPart 
        module. Store the functions in the MbsData instance and assign the pointer
        of functions in the C structure.
            
        """
        self.ptrs_to_symb_fcts["mbs_cons_hJ"] = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : self.__callback_mbs_cons_hJ(self.mbs_cons_hJ,h,Jac))
        self.mbs.mbs_cons_hJ = self.mbs_cons_hJ
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = self.ptrs_to_symb_fcts["mbs_cons_hJ"]
        return
    
    # Callback function for function with advanced arguments
    def __callback_user_cons_hJ(self, fun, h, Jac, tsim):
        __h = np.ctypeslib.as_array(h, (self.mbs.Nuserc+1,))
        __Jac = np.ctypeslib.as_array(Jac[0], (self.mbs.Nuserc+1,self.mbs.njoint+1))
        fun(__h,__Jac,self.mbs,tsim)
        
    def __callback_mbs_cons_hJ(self, fun, h, Jac):
        if __DEBUG__ : print( "DEBUG>>  callback_mbs_cons_hJ")
        __h = np.ctypeslib.as_array(h, (self.mbs.Ncons+1,))
        __Jac = np.ctypeslib.as_array(Jac[0], (self.mbs.Ncons+1,self.mbs.njoint+1))
        fun(__h,__Jac,self.mbs)
        
    def __del__(self):
        libmodules.mbs_delete_part(self.mbs_part_ptr)
        if __DEBUG__ : print("DEBUG>>  MbsPart pointer deleted")
        
    def __unassign_user_fct__(self):
        """ 
        Unassign user function where some args depend on MbsPart module.
        """
        self.ptrs_to_user_fcts = {}
        self.mbs.user_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.user_cons_hJ = user_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("user_cons_hJ"))
        
        return
         
    def __unassign_symbolic_fct__(self):
        """ 
        Unassign symbolic function where some args depend on MbsPart module.
        """
        self.ptrs_to_symb_fcts = {}
        self.mbs.mbs_cons_hJ = None
        self.mbs.mbs_data_ptr.contents.mbs_cons_hJ = mbs_cons_hJ_wrap(lambda h,Jac,mbs,tsim : callback_undefined("mbs_cons_hJ"))
        
        return
    
    def run(self):
        """
        Run a coordinate partitioning.
        
        Options can be setted with the function 'set_options()' 
        Options can be retrieved with the function 'get_options()'            
        """
        # Assing required user functions
        if self.mbs.opt_load_c < 2:
            self.__assign_user_fct__()
            self.mbs.__assign_user_fct__()
        # Assing required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__assign_symbolic_fct__()
            self.mbs.__assign_symb_fct__()
        
        libmodules.mbs_run_part(self.mbs_part_ptr, self.mbs.mbs_data_ptr)
        
        # Unassign functions
        if self.mbs.opt_load_c < 2:
            self.__unassign_user_fct__()
            self.mbs.__unassign_user_fct__()
        # Unassign required symbolic functions
        if self.mbs.opt_load_c < 1:
            self.__unassign_symbolic_fct__()
            self.mbs.__unassign_symb_fct__()
            
    
    def set_options(self, **kwargs):
        """
        Set the specified options for Part module.
        
        Parameters
        ----------
        rowperm : int 
            1 to allow line permutation, 0 to forbid it.
            defaut is 0
        visualise : int 
            Unused but 1 stand for yes, 0 for no.
            defaut is 0
        drivers : int
            Unused but 1 stand for yes, 0 for no.
            defaut is 0
        verbose : int 
            Set to 1 to get messages related to partitioning module. Set to 0 
            to disable message.
            defaut is 1
        clearmbsglobal : int
            Unused, but:
             - 1 is for inout
             - 2 is for out
             - 3 is for none
             - 4 is for all
            defaut is 1
        random_amplitude : float
            Amplitude of the random troubleshoot added to the initial value of 
            generalized coordinates. The randomized values are only used to 
            determine the rank of the Jacobian matrix.
            default is 1.e-2.
        """
        for key, value in kwargs.items():
            if key == "rowperm":
                self.mbs_part_ptr.contents.options.contents.rowperm = value
            elif key == "visualise":
                self.mbs_part_ptr.contents.options.contents.visualise = value
            elif key == "drivers":
                self.mbs_part_ptr.contents.options.contents.drivers = value
            elif key == "verbose":
                self.mbs_part_ptr.contents.options.contents.verbose = value
            elif key == "clearmbsglobal":
                self.mbs_part_ptr.contents.options.contents.clearmbsglobal = value
            elif key == "random_amplitude":
                self.mbs_part_ptr.contents.options.contents.random_amplitude = value
            else:
                print(">>PART>> The option "+key+ " is not defined in this module")

    def get_options(self, *args):
        """
        Get the specified options for Part module.
        
        Parameters
        ----------
        The different options specifed in the documentation of 'set_options()'.
        
        Returns
        -------
        The value of the required options.
        
        """
        options = []
        for key in args:
            if key == "rowperm":
                options.append(self.mbs_part_ptr.contents.options.contents.rowperm)
            elif key == "visualise":
                options.append(self.mbs_part_ptr.contents.options.contents.visualise)
            elif key == "drivers":
                options.append(self.mbs_part_ptr.contents.options.contents.drivers)
            elif key == "verbose":
                options.append(self.mbs_part_ptr.contents.options.contents.verbose)
            elif key == "clearmbsglobal":
                options.append(self.mbs_part_ptr.contents.options.contents.clearmbsglobal)
            elif key == "random_amplitude":
                options.append(self.mbs_part_ptr.contents.options.contents.random_amplitude)
            else:
                print(">>PART>> The option "+key+ " is not defined in this module")
        if len(options) == 0:
            return
        if len(options) == 1:
            return options[0]
        return tuple(options)


    #==========================================================================
    # Defining properties
    #==========================================================================
    @property
    def n_qu(self):
        return self.mbs_part_ptr.contents.n_qu
    
    @property
    def n_qv(self):
        return self.mbs_part_ptr.contents.n_qv
    
    @property
    def n_hu(self):
        return self.mbs_part_ptr.contents.n_hu
    
    @property
    def n_hv(self):
        return self.mbs_part_ptr.contents.n_hv
    
    @property
    def q_closed(self):
        return self.__q_closed
    
    @property
    def mbs_filename(self):
        return self.mbs.mbs_filename
    
    @property
    def project_path(self):
        return self.mbs.project_path
    
    @property
    def mbs_name(self):
        return self.mbs.mbs_name
