# -*- coding: utf-8 -*-
"""
-------------------------------
(c) Universite catholique de Louvain, 2019

Creation : 2019 by O. Lantsoght
Last update : 2019

version MBsysC v1.11.2

-------------------------------
Portable Python interface to MBsysC using Ctypes.

Define the class MbsSensor based on the MbsSensor structure of MBsysC.
This class has the functions required to compute the sensor values.
"""

import ctypes
import numpy as np

# importing libraries
from .._mbsysc_loader.loadlibs import libmodules

# importing MBsysC related classes
from .._mbsysc_loader.forward_decl import MbsSensor_c


#==============================================================================
# Global parameter of the current module
#==============================================================================
__DEBUG__ = False

#==============================================================================
# Defining Python MbsSensor class
#==============================================================================
class MbsSensor(object):
    """
    Sensor are used to compute fields of a point on the multibody system.
    
    Sensor can be computed in pure Python or computed by Python but the memory
    being allocated by the C-library or computed (and allocated) by the c-library.
    
    This class can be used for:
     - Sensor : retrived data of a point of the system
     - GenSensor : retrive data of the origin of any body of the system
     - ForceSensor : retrieve data of a point of the sensor and apply a force on it
        
    Attributes
    ----------
    P : ndarray
        Contains the position of the sensor expressed in the inertial 
        frame: P[1:4]=[P_x, P_y, P_z]
    V : ndarray
        Contains the velocity of the sensor expressed in the inertial 
        frame: V[1:4]=[V_x, V_y, V_z]
    A: ndarray
        Contains the velocity of the sensor expressed in the inertial 
        frame: A[1:4]=[A_x, A_y, A_z]
    R : ndarray
        Contains the rotation matrix from the inertial frame to the sensor
        frame: [X^s]=R.[X^0]
    OM : ndarray
        Contains the angular velocity vector of the sensor expressed in the 
        inertial frame: OM[1:4]=[omega_x, omega_y, omega_z]
    OMP : ndarray
        Contains the angular acceleration vector of the sensor expressed in the 
        inertial frame: OMP[1:4]=[domega_x, domega_y, domega_z]
    J : ndarray
        Contains the jacobian matrix of the sensor: 
        J[1:njoint,1:7]=(d velocity)/(d q)
    id_ : int or None
        The ID of the sensor on the system. It is required to update the sensor
        values without giving the id to the method `sensor` or `gensensor` (both
        having an optionnal parameter `id_`).
    """
    def __init__(self, mbs, id_=None, sensor_ptr=None):
        """ 
        Create an instance of the MbsSensor class.
        
        The sensor can be linked with C-memory (sensor_ptr) or being in pure Python.

        Parameters
        ----------
        mbs : MbsData
            Python MbsData instance of the MBS on wich the sensor is computed.
        id_ : int, optionnal
            ID of the sensor on the MBS.
        sensor_ptr : ctypes.POINTER(MbsSensor_c), optionnal
            A pointer to an existing strucure of MbsSensor_c. If provided the
            Python instance will share its memory with the C structure.
        """
        self.__mbs_data = mbs
        self.id = id_
        
        if (sensor_ptr is None) and (mbs.opt_load_c==0):
            # Sensor is fully computed and allocated in Python
            self.__sensor_ptr = None
            self.__python = True
            
            self.__P = np.array([3,0.,0.,0.])
            self.__R = np.array([[3,0.,0.,0.],
                                 [3,0.,0.,0.],
                                 [3,0.,0.,0.],
                                 [3,0.,0.,0.]])
            self.__V = np.array([3,0.,0.,0.])
            self.__OM = np.array([3,0.,0.,0.])
            self.__A = np.array([3,0.,0.,0.])
            self.__OMP = np.array([3,0.,0.,0.])
            self.__J = np.zeros((6+1,mbs.njoint+1))
            self.__J[0,0] = 6
            self.__J[1:,0] = mbs.njoint
        else:
            # Binding the sensor with a C-memory
            self.__python = False
            if sensor_ptr is None:
                # The sensor must be allocated in the C-library
                self.__sensor_ptr = libmodules.mbs_new_sensor(self.__mbs_data.mbs_data_ptr)
                libmodules.allocate_sensor(self.__sensor_ptr, mbs.njoint)
                libmodules.init_sensor(self.__sensor_ptr, mbs.njoint)
            else:
                # The sensor already exists in the C-library
                self.__sensor_ptr = ctypes.cast(sensor_ptr, ctypes.POINTER(MbsSensor_c))

            self.__P = np.ctypeslib.as_array(self.__sensor_ptr.contents.P, (4,))
            self.__R = np.ctypeslib.as_array(self.__sensor_ptr.contents.R[0], (4,4))
            self.__V = np.ctypeslib.as_array(self.__sensor_ptr.contents.V, (4,))
            self.__OM = np.ctypeslib.as_array(self.__sensor_ptr.contents.OM, (4,))
            self.__A = np.ctypeslib.as_array(self.__sensor_ptr.contents.A, (4,))
            self.__OMP = np.ctypeslib.as_array(self.__sensor_ptr.contents.OMP,(4,))
            self.__J = np.ctypeslib.as_array(self.__sensor_ptr.contents.J[0] , (7,mbs.njoint+1))
            
    def assign_to_C(self,sensor):
        """
        Reassign a python sensor to an existing C MbsSensor structure.
        
        If the sensor was already allocated by the C-library, it will not be
        binded again.

        Parameters
        ----------
        sensor : ctypes.POINTER(MbsSensor_c)
            pointer to the allocated MbsSensor structure
        """
        if self.__python == False:
            print("Sensor is already assigned to a C MbsSensor structure.")
            return
        
        self.__python = False
        sensor = ctypes.cast(sensor, ctypes.POINTER(MbsSensor_c))
        
        self.__P = np.ctypeslib.as_array(sensor.contents.P,   (4,))
        self.__R = np.ctypeslib.as_array(sensor.contents.R[0], (4,4))
        self.__V = np.ctypeslib.as_array(sensor.contents.V,   (4,))
        self.__OM = np.ctypeslib.as_array(sensor.contents.OM,  (4,))
        self.__A = np.ctypeslib.as_array(sensor.contents.A,   (4,))
        self.__OMP = np.ctypeslib.as_array(sensor.contents.OMP, (4,))
        self.__J = np.ctypeslib.as_array(sensor.contents.J[0] , (7,self.mbs.njoint+1))
        
    def sensor(self, id_=None):
        """
        Update the fields of the sensor.
        
        Compute all field of the specified sensor. The sensors have been 
        defined by the user on the MBS in the MBsysPAD interface.

        Parameters
        ----------
        id_ : int, optionnal
            ID of the sensor on the MBS.
        
        Notes
        -----
        If 'id_' is not set on the sensor instance the first sensor is 
        computed. However a warning is generated.
        
        See Also
        --------
        gensensor : compute the sensor located at the origin of a body
        """
        if id_ == None:
            if self.id == None:
                print(">> Warning >> Sensor id was not defined. Using default id (=1)")
                id_ = 1
            else:
                id_ = self.id
        if not self.__mbs_data.mbs_sensor is None:
            self.__mbs_data.mbs_sensor(self, self.__mbs_data, id_)
        else:
            self.__mbs_data.mbs_data_ptr.contents.mbs_sensor(self.__sensor_ptr, self.__mbs_data.mbs_data_ptr, id_)
        
    def gensensor(self, id_=None):
        """
        Update the fields of the sensor.
        
        Compute all field of the specified sensor. The sensors were  automatically
        defined at the origin of each body of the MBS.

        Parameters
        ----------
        id_ : int, optionnal
            ID of the sensor on the MBS. Cannot exceed the number of bodies.
        
        Notes
        -----
        If 'id_' is not set on the sensor instance, the first body is considered.
        However a warning is generated.
        
        See Also
        --------
        sensor : compute the user-specified sensor of the MBS.
        """
        if id_ == None:
            if self.id == None:
                print(">> Warning >> Body id was not defined for gensensor."
                      "Using default id (=1).")
                id_ = 1
            else:
                id_ = self.id
            
        self.__mbs_data.mbs_gensensor(self, self.__mbs_data, id_)
        
    
    @property
    def P(self):
        """See attributes 'P'"""
        return self.__P
        
    @property
    def R(self):
        """See attributes 'R'"""
        return self.__R
        
    @property
    def V(self):
        """See attributes 'V'"""
        return self.__V
        
    @property
    def OM(self):
        """See attributes 'OM'"""
        return self.__OM
        
    @property
    def A(self):
        """See attributes 'A'"""
        return self.__A
        
    @property
    def OMP(self):
        """See attributes 'OMP'"""
        return self.__OMP
        
    @property
    def J(self):
        """See attributes 'J'"""
        return self.__J

